from unittest import TestCase
from unittest.mock import patch

from core.basic_models.classifiers.basic_classifiers import ExternalClassifier
from scenarios.scenario_models.field.field_filler_description import ClassifierFiller, ClassifierFillerMeta
from smart_kit.utils.picklable_mock import PicklableMock


class TestClassifierFiller(TestCase):

    def setUp(self):
        test_items = {
            "type": "classifier",
            "intents": ["да", "нет"],
            "classifier": {"type": "external", "classifier": "hello_scenario_classifier"}
        }
        self.filler = ClassifierFiller(test_items)

        self.mock_text_preprocessing_result = PicklableMock()
        self.mock_user = PicklableMock()
        self.mock_user.descriptions = {
            "external_classifiers": ["read_book_or_not_classifier", "hello_scenario_classifier"]}

    @patch.object(
        ExternalClassifier,
        "find_best_answer",
        return_value=[{"answer": "нет", "score": 0.7, "other": False}, {"answer": "да", "score": 0.3, "other": False}]
    )
    def test_filler_extract(self, mock_classifier_model):
        """Тест кейз проверяет что поле заполнено наиболее вероятным значением, которое вернула модель."""
        expected_res = "нет"
        actual_res = self.filler.extract(self.mock_text_preprocessing_result, self.mock_user)
        self.assertEqual(expected_res, actual_res)

    @patch.object(ExternalClassifier, "find_best_answer", return_value=[])
    def test_filler_extract_if_no_model_answer(self, mock_classifier_model):
        """Тест кейз проверяет что поле осталось не заполненным те результат None, если модель не выдала ответ."""
        actual_res = self.filler.extract(self.mock_text_preprocessing_result, self.mock_user)
        self.assertIsNone(actual_res)


class TestClassifierFillerMeta(TestCase):

    def setUp(self):
        test_items = {
            "type": "classifier_meta",
            "intents": ["да", "нет"],
            "classifier": {"type": "external", "classifier": "hello_scenario_classifier"}
        }
        self.filler_meta = ClassifierFillerMeta(test_items)

        self.mock_text_preprocessing_result = PicklableMock()
        self.mock_user = PicklableMock()
        self.mock_user.descriptions = {
            "external_classifiers": ["read_book_or_not_classifier", "hello_scenario_classifier"]}

    @patch.object(ExternalClassifier, "find_best_answer", return_value=[{"answer": "нет", "score": 1.0, "other": False}])
    def test_filler_extract(self, mock_classifier_model):
        """Тест кейз проверяет что мы получаем тот же самый ответ, что вернула модель."""
        expected_res = [{"answer": "нет", "score": 1.0, "other": False}]
        actual_res = self.filler_meta.extract(self.mock_text_preprocessing_result, self.mock_user)
        self.assertEqual(expected_res, actual_res)

    @patch.object(ExternalClassifier, "find_best_answer", return_value=[])
    def test_filler_extract_if_no_model_answer(self, mock_classifier_model):
        """Тест кейз проверяет результат None, если модель не выдала ответ."""
        actual_res = self.filler_meta.extract(self.mock_text_preprocessing_result, self.mock_user)
        self.assertIsNone(actual_res)



# coding=utf-8
from typing import Iterable, Dict, Optional, Set, Any, List, Union, Tuple
import json
import uuid

from core.configs.global_constants import CALLBACK_ID_HEADER
from core.message.app_info import AppInfo
from core.message.device import Device
from core.names import field
import core.logging.logger_constants as log_const
from core.logging.logger_utils import log
from core.utils.masking_message import masking
from core.utils.utils import current_time_ms
from core.message.msg_validator import MessageValidator

from smart_kit.configs import get_app_config


class Headers:
    def __init__(self, data):
        self.raw = dict(data)

    def __getitem__(self, item):
        return self.raw[item].decode()

    def __contains__(self, item):
        return item in self.raw

    def get(self, key, default=None, encoding="utf-8"):
        res = self.raw.get(key)
        if res is None:
            return default
        return res.decode(encoding=encoding)

    def __bool__(self):
        return bool(self.raw)


class SmartAppFromMessage:
    MESSAGE_NAME = "messageName"
    MESSAGE_ID = "messageId"
    UUID = "uuid"
    PAYLOAD = "payload"
    SESSION_ID = "sessionId"
    CALLBACK_ID_HEADER_NAME = CALLBACK_ID_HEADER

    # following fields are used for validation
    incremental_id: int
    message_name: str
    payload: dict
    uuid: dict

    def __init__(self, value: Dict[str, Any], topic_key: str = None, creation_time: Optional[int] = None,
                 kafka_key: Optional[str] = None, headers: Optional[Iterable[Tuple[Any, Any]]] = None,
                 masking_fields: Optional[Union[Dict[str, int], List[str]]] = None, headers_required: bool = True,
                 validators: Iterable[MessageValidator] = ()):
        self.logging_uuid = str(uuid.uuid4())
        self._value = value
        self.topic_key = topic_key
        self.kafka_key = kafka_key
        self.creation_time = creation_time or current_time_ms()
        self._headers_required = headers_required
        if self._headers_required and headers is None:
            raise LookupError(f"{self.__class__.__name__} no incoming headers.")
        self.headers = Headers(headers or {})
        self._callback_id = None  # FIXME: by some reason it possibly to change callback_id
        self.masking_fields = masking_fields
        self.validators = validators

    def validate(self) -> bool:
        """
            Try to json.load message and check for all required fields
        """

        for validator in self.validators:
            if not validator.validate(self.message_name, self.payload):
                return False

        if self._headers_required and not self.headers:
            log("Message headers is empty", level="ERROR")
            return False

        try:
            for r_field in self._required_fields:
                if r_field not in self.as_dict:
                    self.print_validation_error(r_field)
                    return False

                if r_field not in self.__annotations__:
                    continue

                if not isinstance(
                        self.as_dict[r_field],
                        self.__annotations__[r_field],
                ):
                    self.print_validation_error(
                        r_field,
                        self.__annotations__[r_field],
                    )
                    return False

        except (json.JSONDecodeError, TypeError):
            log(
                "Message validation error: json decode error",
                exc_info=True,
                level="ERROR",
            )
            self.print_validation_error()
            return False

        return True

    def print_validation_error(
            self,
            required_field: Optional[str] = None,
            required_field_type: Optional[str] = None,
    ) -> None:
        if self._value:
            params = {
                "value": str(self._value),
                "required_field": required_field,
                "required_field_type": required_field_type,
                log_const.KEY_NAME: log_const.EXCEPTION_VALUE
            }
            if required_field and required_field_type:
                log(
                    "Message validation error: Expected '%(required_field)s'"
                    " of type '%(required_field_type)s': %(value)s",
                    params=params,
                    level="ERROR",
                )
            elif required_field:
                log(
                    "Message validation error: Required field "
                    "'%(required_field)s' is missing: %(value)s",
                    params=params,
                    level="ERROR",
                )
            else:
                log(
                    "Message validation error: Format is wrong: %(value)s",
                    params=params,
                    level="ERROR",
                )
        else:
            log("Message validation error: Message is empty", level="ERROR")

    @property
    def _callback_id_header_name(self) -> str:
        return CALLBACK_ID_HEADER

    @property
    def _required_fields(self) -> Set[str]:
        return {self.MESSAGE_ID, self.UUID, self.PAYLOAD, self.SESSION_ID, self.MESSAGE_NAME}

    @property
    def session_id(self) -> Optional[str]:
        return self.as_dict.get(self.SESSION_ID)

    # database user_id
    @property
    def db_uid(self) -> str:
        return "{}_{}".format(self.uid, self.channel)

    @property
    def channel(self) -> Optional[str]:
        return self.uuid.get(field.USER_CHANNEL)

    @property
    def uid(self) -> Optional[str]:
        return self.uuid.get(field.USER_ID)

    @property
    def sub(self) -> Optional[str]:
        return self.uuid.get(field.SUB)

    @property
    def uuid(self) -> Dict[str, str]:
        return self.as_dict[self.UUID]

    @property
    def payload(self) -> Dict[str, Any]:
        return self.as_dict[self.PAYLOAD]

    @payload.setter
    def payload(self, payload):
        self._value[self.PAYLOAD] = payload

    @property
    def type(self) -> str:
        return self.as_dict[self.MESSAGE_NAME]

    def project_name(self) -> Optional[str]:
        return self.payload.get(field.PROJECT_NAME)

    @property
    def intent(self) -> Optional[str]:
        return self.payload.get(field.INTENT)

    @property
    def device(self) -> Device:
        return Device(self.payload.get(field.DEVICE) or {})

    @property
    def app_info(self) -> AppInfo:
        return AppInfo(self.payload.get(field.APP_INFO) or {})

    @property
    def smart_bio(self) -> Dict[str, Any]:
        return self.payload.get(field.SMART_BIO) or {}

    @property
    def annotations(self) -> Dict[str, Dict[str, float]]:
        annotations = self.payload.get(field.ANNOTATIONS) or {}
        for annotation in annotations:
            classes = annotations[annotation][field.CLASSES]
            probas = annotations[annotation][field.PROBAS]
            annotations[annotation] = dict(zip(classes, probas))
        return annotations

    @property
    def callback_id(self) -> Optional[str]:
        if self._callback_id is not None:
            return self._callback_id

        try:
            from smart_kit.start_points.main_loop_http import HttpMainLoop
            if issubclass(get_app_config().MAIN_LOOP, HttpMainLoop):
                return str(self.incremental_id)
            return self.headers[self._callback_id_header_name]
        except KeyError:
            log(f"{self._callback_id_header_name} missed in headers for message_id %(message_id)s",
                params={log_const.KEY_NAME: "callback_id_missing", "message_id": self.incremental_id}, level="WARNING")
            return None

    @callback_id.setter
    def callback_id(self, value: str):
        self._callback_id = value

    # noinspection PyMethodMayBeStatic
    def generate_new_callback_id(self) -> str:
        from smart_kit.start_points.main_loop_http import HttpMainLoop
        if issubclass(get_app_config().MAIN_LOOP, HttpMainLoop):
            return str(self.incremental_id)
        return str(uuid.uuid4())

    @property
    def masked_value(self) -> str:
        masked_data = masking(self.as_dict, self.masking_fields)
        return json.dumps(masked_data, ensure_ascii=False)

    @property
    def message_name(self) -> str:
        return self.as_dict[self.MESSAGE_NAME]

    @message_name.setter
    def message_name(self, message_name: str):
        self._value[self.MESSAGE_NAME] = message_name

    # unique message_id
    @property
    def incremental_id(self) -> int:
        return self.as_dict[self.MESSAGE_ID]

    @property
    def as_dict(self) -> Dict[str, Any]:
        return self._value

    @property
    def as_str(self) -> str:
        return json.dumps(self._value, ensure_ascii=False)


basic_error_message = SmartAppFromMessage(
    {
        "messageName": "ERROR",
        "messageId": -1,
        "uuid": -1,
        "payload": {},
        "sessionId": -1
    },
    headers={},
    headers_required=False,
)



# coding=utf-8

from typing import Type, Iterable
import signal

import scenarios.logging.logger_constants as log_const
from core.db_adapter.db_adapter import DBAdapterException
from core.db_adapter.db_adapter import db_adapter_factory
from core.logging.logger_utils import log
from core.message.from_message import SmartAppFromMessage
from core.monitoring.monitoring import monitoring
from core.monitoring.healthcheck_handler import RootResource
from core.monitoring.twisted_server import TwistedServer
from core.model.base_user import BaseUser
from core.basic_models.parametrizers.parametrizer import BasicParametrizer
from core.message.msg_validator import MessageValidator
from smart_kit.start_points.postprocess import PostprocessMainLoop
from smart_kit.models.smartapp_model import SmartAppModel


class BaseMainLoop:

    def __init__(
            self, model: SmartAppModel,
            user_cls: Type[BaseUser],
            parametrizer_cls: Type[BasicParametrizer],
            postprocessor_cls: Type[PostprocessMainLoop],
            settings,
            to_msg_validators: Iterable[MessageValidator] = (),
            from_msg_validators: Iterable[MessageValidator] = (),
            *args, **kwargs
    ):
        log("%(class_name)s.__init__ started.", params={
            log_const.KEY_NAME: log_const.STARTUP_VALUE,
            "class_name": self.__class__.__name__
        })
        try:
            signal.signal(signal.SIGINT, self.stop)
            signal.signal(signal.SIGTERM, self.stop)
            self.settings = settings
            self.app_name = self.settings.app_name
            self.model: SmartAppModel = model
            self.user_cls = user_cls
            self.parametrizer_cls = parametrizer_cls
            self.postprocessor = postprocessor_cls()
            self.db_adapter = self.get_db()
            self.is_work = True
            self.to_msg_validators: Iterable[MessageValidator] = to_msg_validators
            self.from_msg_validators: Iterable[MessageValidator] = from_msg_validators

            template_settings = self.settings["template_settings"]

            save_tries = template_settings.get("user_save_collisions_tries", 0)

            self.user_save_check_for_collisions = True if save_tries > 0 else False
            self.user_save_collisions_tries = max(save_tries, 1)

            self.health_check_server = self._create_health_check_server(template_settings)
            self._init_monitoring_config(template_settings)

            log("%(class_name)s.__init__ completed.", params={log_const.KEY_NAME: log_const.STARTUP_VALUE,
                                                                        "class_name": self.__class__.__name__})
        except:
            log("%(class_name)s.__init__ exception.", params={log_const.KEY_NAME: log_const.STARTUP_VALUE,
                                                                        "class_name": self.__class__.__name__},
                          level="ERROR", exc_info=True)
            raise

    def get_db(self):
        db_adapter = db_adapter_factory(self.settings["template_settings"].get("db_adapter", {}))
        if db_adapter.IS_ASYNC:
            raise Exception(
                f"Async adapter {db_adapter.__class__.__name__} doesnt compare with {self.__class__.__name__}"
            )
        db_adapter.connect()
        return db_adapter

    def _generate_answers(self, user, commands, message, **kwargs):
        raise NotImplementedError

    def _create_health_check_server(self, settings):
        health_check_server = None
        if settings["health_check"].get("enabled"):
            log("Init health_check started.", params={log_const.KEY_NAME: log_const.STARTUP_VALUE})
            health_check = settings["health_check"]
            health_check_server = TwistedServer(
                health_check["port"],
                health_check["interface"],
                RootResource,
                settings["environment"] in health_check.get("debug_envs", [])
            )
        return health_check_server

    def _init_monitoring_config(self, template_settings):
        monitoring_config = template_settings["monitoring"]
        monitoring.apply_config(monitoring_config)
        monitoring.init_metrics(app_name=self.app_name)

    def load_user(self, db_uid, message: SmartAppFromMessage):
        db_data = None
        load_error = False
        try:
            db_data = self.db_adapter.get(db_uid)
        except (DBAdapterException, ValueError):
            log("Failed to get user data", params={log_const.KEY_NAME: log_const.FAILED_DB_INTERACTION,
                                                   log_const.REQUEST_VALUE: message.as_str}, level="ERROR")
            load_error = True
            monitoring.counter_load_error(self.app_name)
        return self.user_cls(
            message.uid,
            message=message,
            db_data=db_data,
            settings=self.settings,
            descriptions=self.model.scenario_descriptions,
            parametrizer_cls=self.parametrizer_cls,
            load_error=load_error
        )

    def save_user(self, db_uid, user, message: SmartAppFromMessage):
        no_collisions = True
        if user.do_not_save:
            log("User %(uid)s will not saved", user=user, params={"uid": user.id,
                                                                  log_const.KEY_NAME: "user_will_not_saved"})
        else:

            no_collisions = True
            try:
                str_data = user.raw_str
                if user.initial_db_data and self.user_save_check_for_collisions:
                    no_collisions = self.db_adapter.replace_if_equals(db_uid,
                                                                      sample=user.initial_db_data,
                                                                      data=str_data)
                else:
                    self.db_adapter.save(db_uid, str_data)
            except (DBAdapterException, ValueError):
                log("Failed to set user data", params={log_const.KEY_NAME: log_const.FAILED_DB_INTERACTION,
                                                       log_const.REQUEST_VALUE: message.as_str}, level="ERROR")
                monitoring.counter_save_error(self.app_name)
            if not no_collisions:
                monitoring.counter_save_collision(self.app_name)
        return no_collisions

    def run(self):
        raise NotImplementedError

    def stop(self, signum, frame):
        raise NotImplementedError



# coding: utf-8
import unittest

from smart_kit.configs import logger_config


class ConfigsTest1(unittest.TestCase):
    def test_LoggerConfig(self):
        obj = logger_config.LoggerConfig("./testing/data")
        self.assertTrue(obj.config_path == "./testing/data")
        self.assertTrue(obj.repositories[0].key == 'logging_config')
        self.assertTrue(obj.repositories[0].filename == "./testing/data/logging_config.yml")
        self.assertTrue(obj._subfolder == "./testing/data")



# coding: utf-8
import sys
import traceback

from core.descriptions.descriptions import Descriptions
from core.logging.logger_utils import log
from core.utils.exception_handlers import exc_handler

import scenarios.logging.logger_constants as log_const
from smart_kit.handlers.handle_close_app import HandlerCloseApp
from smart_kit.handlers.handler_take_profile_data import HandlerTakeProfileData
from smart_kit.names.message_names import MESSAGE_TO_SKILL, LOCAL_TIMEOUT, RUN_APP, SERVER_ACTION, CLOSE_APP, \
    TAKE_PROFILE_DATA
from smart_kit.handlers.handler_run_app import HandlerRunApp
from smart_kit.handlers.handle_respond import HandlerRespond
from smart_kit.handlers.handler_text import HandlerText
from smart_kit.handlers.handler_timeout import HandlerTimeout
from smart_kit.handlers.handle_server_action import HandlerServerAction
from smart_kit.resources import SmartAppResources
from smart_kit.utils import get_callback_action_params, set_debug_info
from core.monitoring.monitoring import monitoring


class SmartAppModel:
    # additional_handlers format:
    # {"MESSAGE_NAME": {"handler": HandlerText, "params": {"dialogue_manager": custom_dialogue_manager}}}
    # "params" are passed as kwargs to handler.__init__(), optional
    additional_handlers = {}

    def __init__(self, resources: SmartAppResources, dialogue_manager_cls, custom_settings, **kwargs):
        log(
            f"{self.__class__.__name__}.__init__ started.", params={log_const.KEY_NAME: log_const.STARTUP_VALUE}
        )
        self.resources = resources
        self.scenario_descriptions = Descriptions(self.resources.registered_repositories)
        self.template_settings = custom_settings["template_settings"]
        self.app_name = custom_settings.app_name
        self.dialogue_manager = dialogue_manager_cls(scenario_descriptions=self.scenario_descriptions,
                                                     app_name=self.app_name)

        self._handlers = {
            MESSAGE_TO_SKILL: HandlerText(self.app_name, dialogue_manager=self.dialogue_manager),
            RUN_APP: HandlerRunApp(self.app_name, dialogue_manager=self.dialogue_manager),
            LOCAL_TIMEOUT: HandlerTimeout(self.app_name),
            SERVER_ACTION: HandlerServerAction(self.app_name),
            CLOSE_APP: HandlerCloseApp(self.app_name),
            TAKE_PROFILE_DATA: HandlerTakeProfileData(self.app_name),
        }
        self._handlers.update({
            message_name: HandlerRespond(self.app_name, action_name=action_name)
            for message_name, action_name in self.resources.get("responses", {}).items()
        })
        self.init_additional_handlers()

        log(
            f"{self.__class__.__name__}.__init__ finished.", params={log_const.KEY_NAME: log_const.STARTUP_VALUE}
        )

    def get_handler(self, message_type):
        return self._handlers[message_type]

    def init_additional_handlers(self):
        self._handlers.update({
            message_name: handler_dict["handler"](self.app_name, **handler_dict.get("params", {}))
            for message_name, handler_dict in self.additional_handlers.items()
        })

    @exc_handler(on_error_obj_method_name="on_answer_error")
    def answer(self, message, user):
        user.expire()
        handler = self.get_handler(message.type)

        if not user.load_error:
            commands = handler.run(message.payload, user)
        else:
            log("Error in loading user data", user, level="ERROR", exc_info=True)
            raise Exception("Error in loading user data")

        return commands

    def on_answer_error(self, message, user):
        user.do_not_save = True
        monitoring.counter_exception(self.app_name)
        params = {log_const.KEY_NAME: log_const.DIALOG_ERROR_VALUE,
                  "message_id": user.message.incremental_id,
                  "masked_message": user.message.masked_value}

        log("exc_handler: Failed to process message. Exception occurred. Fail in MESSAGE: %(masked_message)s",
            user, params, level="ERROR", exc_info=True)

        callback_action_params = get_callback_action_params(user)
        exc_type, exc_value, exc_traceback = sys.exc_info()
        error = '\n'.join(traceback.format_exception(exc_type, exc_value, exc_traceback))
        if user.settings["template_settings"].get("debug_info"):
            set_debug_info(self.app_name, callback_action_params, error)
        exception_action = user.descriptions["external_actions"]["exception_action"]
        commands = exception_action.run(user=user, text_preprocessing_result=None,
                                        params=callback_action_params)
        return commands



import re


class Num2Text:
    """
    конвертация чисел в числительные, основано на https://github.com/seriyps/ru_number_to_text
    доделана работа с дробной частью, добавлено больше порядков
    числа по модулю должны быть меньше 1000 дециллионов, дробная часть округляется до 6 порядка
    можно выбирать мужской (m) или женский (f) род (один / одна)
    """
    minus = 'минус'
    units = ('ноль', ('один', 'одна'), ('два', 'две'), 'три', 'четыре', 'пять', 'шесть', 'семь', 'восемь', 'девять')
    teens = ('десять', 'одиннадцать', 'двенадцать', 'тринадцать', 'четырнадцать', 'пятнадцать', 'шестнадцать',
             'семнадцать', 'восемнадцать', 'девятнадцать')
    tens = (teens, 'двадцать', 'тридцать', 'сорок', 'пятьдесят', 'шестьдесят', 'семьдесят', 'восемьдесят', 'девяносто')
    hundreds = ('сто', 'двести', 'триста', 'четыреста', 'пятьсот', 'шестьсот', 'семьсот', 'восемьсот', 'девятьсот')
    orders = (
        (('тысяча', 'тысячи', 'тысяч'), 'f'),
        (('миллион', 'миллиона', 'миллионов'), 'm'),
        (('миллиард', 'миллиарда', 'миллиардов'), 'm'),
        (('триллион', 'триллиона', 'триллионов'), 'm'),
        (('квадриллион', 'квадриллиона', 'квадриллионов'), 'm'),
        (('квинтиллион', 'квинтиллиона', 'квинтиллионов'), 'm'),
        (('секстиллион', 'секстиллиона', 'секстиллионов'), 'm')
    )
    fractionals = (
        ('десятая', 'десятых'), ('сотая', 'сотых'), ('тысячная', 'тысячных'), ('десятитысячная', 'десятитысячных'),
        ('стотысячная', 'стотысячных'), ('милионная', 'милионных')
    )
    fractionals_num = len(fractionals)

    def _thousand(self, rest, sex):
        # Converts numbers from 19 to 999
        prev = 0
        plural = 2
        name = []
        use_teens = 10 <= rest % 100 <= 19
        if not use_teens:
            data = ((self.units, 10), (self.tens, 100), (self.hundreds, 1000))
        else:
            data = ((self.teens, 10), (self.hundreds, 1000))
        for names, x in data:
            cur = int(((rest - prev) % x) * 10 / x)
            prev = rest % x
            if x == 10 and use_teens:
                plural = 2
                name.append(self.teens[cur])
            elif cur == 0:
                continue
            elif x == 10:
                name_ = names[cur]
                if isinstance(name_, tuple):
                    name_ = name_[0 if sex == 'm' else 1]
                name.append(name_)
                if 2 <= cur <= 4:
                    plural = 1
                elif cur == 1:
                    plural = 0
                else:
                    plural = 2
            else:
                name.append(names[cur - 1])
        return plural, name

    def int2text(self, num: int, sex='m'):
        if num == 0:
            return self.units[0]
        main_units = (('', '', ''), sex)
        _orders = (main_units,) + self.orders
        rest = abs(num)
        ord = 0
        name = []
        while rest > 0:
            plural, nme = self._thousand(rest % 1000, _orders[ord][1])
            if nme or ord == 0:
                name.append(_orders[ord][0][plural])
            name += nme
            rest = rest // 1000
            ord += 1
        if num < 0:
            name.append(self.minus)
        name.reverse()
        return ' '.join(name).strip()

    def replace_everything_in_text(self, text, sex='m'):
        return re.sub(
            pattern="[+-]?\d+(?:\.\d+)?",
            repl=lambda x: " {} ".format(self(float(x.group()), sex=sex)),
            string=" {} ".format(text)
        ).strip().replace("  ", " ")

    def __call__(self, num, sex='m'):
        num = float(num)
        num = round(num, self.fractionals_num)
        int_num = int(num)
        if int_num == num:
            return self.int2text(num, sex)
        frac = int(str(round(abs(num) - abs(int_num), self.fractionals_num))[2:])
        if frac == 0:
            return self.int2text(round(num), sex)
        if frac == 5:
            return self.int2text(int_num, sex) + " с половиной"
        return "{} {} {} {}".format(
            self.int2text(int_num, 'f'),
            "целая" if (int_num % 10 == 1) and (int_num % 100 != 11) else "целых",
            self.int2text(frac, 'f'),
            self.fractionals[len(str(frac)) - 1][0 if ((frac % 10 == 1) and (frac % 100 != 11)) else 1]
        )


if __name__ == '__main__':
    num2text = Num2Text()
    # demo
    t = "На улице -2 градуса. Битрикс24. Круассанам 7дней"
    print("Проверим режим замены вхождений по всему тексту: \"{}\"".format(t))
    print(num2text.replace_everything_in_text(t))
    print()

    for t in [11, 167342, 17.5, 12.34, -2.67]:
        r = num2text(t, 'm')
        print(">", t)
        print(r)
    print(">", 22)
    print(num2text(22, 'm'))
    print(">", 22)
    print(num2text(22, 'f'))
    # interactive
    while True:
        t = float(input("> "))
        r = num2text(t, 'm')
        print(r)



# coding: utf-8
from collections import defaultdict

from core.model.lazy_items import LazyItems


class LoopException(Exception):
    pass


class RegisteredScenario:
    def __init__(self):
        self._values = dict()

    def __setitem__(self, key, value):
        self._values[key] = value

    def __getitem__(self, key):
        value = self._values.get(key)
        if value is None:
            value = BaseScenarioModel
        return value


scenario_models = RegisteredScenario()


def scenario_model_factory(value, description, user):
    _type = type(description)
    model = scenario_models[_type]
    return model(value, description, user)


class BaseScenarioModel:
    def __init__(self, items, description, user):
        self.description = description

    def set_break(self):
        pass

    @property
    def break_scenario(self):
        return False

    @property
    def raw(self):
        return None


class ScenarioModels(LazyItems):
    def __init__(self, items, descriptions, user):
        super(ScenarioModels, self).__init__(items, descriptions, user, scenario_model_factory)


class MetaClassifierScenarioModel(BaseScenarioModel):
    def __init__(self, value, description, user):
        super(MetaClassifierScenarioModel, self).__init__(value, description, user)
        self.scenario_id = value


class FormFillingScenarioModel(BaseScenarioModel):
    def __init__(self, value, description, user):
        super(FormFillingScenarioModel, self).__init__(value, description, user)
        self._break_scenario = False
        self._shared_data = {}
        self._postprocessing_data = list()

    def set_break(self):
        self._break_scenario = True

    def set_shared(self, key, value):
        self._shared_data[key] = value

    def get_shared(self, key):
        return self._shared_data.get(key)

    @property
    def break_scenario(self):
        return self._break_scenario

    @property
    def postprocessing_data(self):
        return self._postprocessing_data


class TreeScenarioModel(FormFillingScenarioModel):

    DEFAULT_COUNT = 10

    def __init__(self, items, description, user):
        super(TreeScenarioModel, self).__init__(items, description, user)
        items = items or {}
        self.current_node = items.get("current_node")
        self.number_of_calls = defaultdict(int)

    def add_count(self, node_id):
        self.number_of_calls[node_id] += 1
        if self.number_of_calls[node_id] == self.DEFAULT_COUNT:
            raise LoopException("the scenario is looped")

    @property
    def raw(self):
        if self.current_node:
            return {"current_node": self.current_node}



from core.descriptions.smart_updatable_descriptions_items import SmartUpdatableDescriptionsItems
from scenarios.scenario_models.field.field_filler_description import field_filler_factory


class ExternalFieldFillerDescriptions(SmartUpdatableDescriptionsItems):
    def __init__(self, items):
        super(ExternalFieldFillerDescriptions, self).__init__(field_filler_factory, items, ordered=True)



import io
import os
import fnmatch

from core.db_adapter.db_adapter import DBAdapter
from core.db_adapter import error


class OSAdapter(DBAdapter):
    def _save(self, id, data):
        raise error.NotSupportedOperation

    def _replace_if_equals(self, id, sample, data):
        raise error.NotSupportedOperation

    def _get(self, id):
        raise error.NotSupportedOperation

    def connect(self):
        pass

    def _on_prepare(self):
        pass

    @property
    def source(self):
        return self

    def _list_dir(self, path):
        result = []
        for path, subdirs, files in os.walk(path):
            result.extend([os.path.join(path, name) for name in files if not name.startswith(".")])
        return result

    def _open(self, filename, *args, **kwargs):
        return io.open(filename, *args, **kwargs)

    def _get_counter_name(self):
        return "os_adapter"

    def _glob(self, path, pattern):
        files_list = self._list_dir(path)
        filtered = fnmatch.filter(files_list, pattern)
        return filtered

    def _path_exists(self, path):
        return os.path.exists(path)

    def _mtime(self, path):
        return os.path.getmtime(path)



# coding: utf-8
import unittest

from core.basic_models.variables.variables import Variables


class VariablesTest(unittest.TestCase):
    def setUp(self) -> None:
        self.variables = Variables(None, None)

    def test_set(self):
        self.variables.set("key", "value")
        self.assertEqual(self.variables.values, {"key": "value"})

    def test_get(self):
        self.variables.set("key", "value")
        self.assertEqual(self.variables.get("key"), "value")
        self.assertEqual(self.variables.get("nonexistkey", "defaultvalue"), "defaultvalue")

    def test_update(self):
        self.variables.set("key", "value")
        self.variables.update("key", "newvalue")
        self.assertEqual(self.variables.get("key"), "newvalue")

    def test_delete(self):
        self.variables.set("key", "value")
        self.variables.delete("key")
        self.assertEqual(self.variables.values, {})

    def test_expire(self):
        self.variables.set("key", "value", ttl=0)
        self.assertEqual(self.variables.values, {})

    def test_clear(self):
        self.variables.set("key_1", "value_1")
        self.variables.set("key_2", "value_2")
        self.variables.clear()
        self.assertEqual(self.variables.values, {})



# coding: utf-8
import unittest
from unittest.mock import Mock

from scenarios.requirements.requirements import TemplateInArrayRequirement, ArrayItemInTemplateRequirement, \
    RegexpInTemplateRequirement
from smart_kit.utils.picklable_mock import PicklableMock


class MockRequirement:
    def __init__(self, items=None):
        items = items or {}
        self.cond = items.get("cond") or False

    def check(self, text_preprocessing_result, user):
        return self.cond


class MockTextNormalizationResult:
    def __init__(self, normalized=None, number_of_numbers=None, currencies_number=None, tokens=None):
        number_of_numbers = number_of_numbers or 0
        currencies_number = currencies_number or 0
        if normalized:
            self.words_tokenized = list(token.get("text") for token in normalized)
            self.words_tokenized_set = set(self.words_tokenized)
        else:
            self.words_tokenized = list()
            self.words_tokenized_set = set()
        self.number_of_numbers = number_of_numbers
        self.currencies_number = currencies_number
        self.tokenized_elements_list = tokens


class MockAmountOperator:
    def __init__(self, items):
        self.amount = items["amount"]

    def compare(self, value):
        return value > self.amount


class MockOperator:
    def __init__(self, amount):
        self.amount = amount

    def compare(self, value):
        return value > self.amount


class EQMockOperator:
    def __init__(self, amount):
        self.amount = amount

    def compare(self, value):
        return value == self.amount


class RequirementTest(unittest.TestCase):

    def test_template_in_array_req_true(self):
        items = {
            "template": "{{ payload.userInfo.tbcode }}",
            "items": ["32", "33"]
        }
        requirement = TemplateInArrayRequirement(items)
        params = {"payload": {
            "userInfo": {
                "tbcode": "32"
            },
            "message": "BBB"
        }}
        user = PicklableMock()
        user.parametrizer = PicklableMock()
        user.parametrizer.collect = Mock(return_value=params)
        self.assertTrue(requirement.check(None, user))

    def test_template_in_array_req_true2(self):
            items = {
                "template": "{{ payload.message.strip() }}",
                "items": ["AAA", "BBB", "CCC"]
            }
            requirement = TemplateInArrayRequirement(items)
            params = {"payload": {
                "userInfo": {
                    "tbcode": "32"
                },
                "message": " BBB    "
            }}
            user = PicklableMock()
            user.parametrizer = PicklableMock()
            user.parametrizer.collect = Mock(return_value=params)
            self.assertTrue(requirement.check(None, user))

    def test_template_in_array_req_false(self):
            items = {
                "template": "{{ payload.message.strip() }}",
                "items": ["AAA", "CCC"]
            }
            requirement = TemplateInArrayRequirement(items)
            params = {"payload": {
                "userInfo": {
                    "tbcode": "32",
                },
                "message": " BBB "
            }}
            user = PicklableMock()
            user.parametrizer = PicklableMock()
            user.parametrizer.collect = Mock(return_value=params)
            self.assertFalse(requirement.check(None, user))


    def test_array_in_template_req_true(self):
            items = {
                "template": {
                    "type": "unified_template",
                    "template": "{{ payload.userInfo.departcode.split('/')|tojson }}",
                    "loader": "json"
                },
                "items": ["111", "456"]
            }
            requirement = ArrayItemInTemplateRequirement(items)
            params = {"payload": {
                "userInfo": {
                    "tbcode": "32",
                    "departcode": "123/2345/456"
                },
                "message": " BBB "
            }}
            user = PicklableMock()
            user.parametrizer = PicklableMock()
            user.parametrizer.collect = Mock(return_value=params)
            self.assertTrue(requirement.check(None, user))


    def test_array_in_template_req_true2(self):
            items = {
                "template": "{{ payload.message.strip() }}",
                "items": ["AAA", "BBB"]
            }
            requirement = ArrayItemInTemplateRequirement(items)
            params = {"payload": {
                "userInfo": {
                    "tbcode": "32",
                    "departcode": "123/2345/456"
                },
                "message": " BBB "
            }}
            user = PicklableMock()
            user.parametrizer = PicklableMock()
            user.parametrizer.collect = Mock(return_value=params)
            self.assertTrue(requirement.check(None, user))


    def test_array_in_template_req_false(self):
            items = {
                "template": {
                    "type": "unified_template",
                    "template": "{{ payload.userInfo.departcode.split('/')|tojson }}",
                    "loader": "json"
                },
                "items": ["111", "222"]
            }
            requirement = ArrayItemInTemplateRequirement(items)
            params = {"payload": {
                "userInfo": {
                    "tbcode": "32",
                    "departcode": "123/2345/456"
                },
                "message": " BBB "
            }}
            user = PicklableMock()
            user.parametrizer = PicklableMock()
            user.parametrizer.collect = Mock(return_value=params)
            self.assertFalse(requirement.check(None, user))


    def test_regexp_in_template_req_true(self):
            items = {
                "template": "{{ payload.message.strip() }}",
                "regexp": "(^|\s)[Фф](\.|-)?1(\-)?(у|У)?($|\s)"
            }
            requirement = RegexpInTemplateRequirement(items)
            params = {"payload": {
                "userInfo": {
                    "tbcode": "32",
                },
                "message": "карточки ф1у"
            }}
            user = PicklableMock()
            user.parametrizer = PicklableMock()
            user.parametrizer.collect = Mock(return_value=params)
            self.assertTrue(requirement.check(None, user))


    def test_regexp_in_template_req_false(self):
            items = {
                "template": "{{ payload.message.strip() }}",
                "regexp": "(^|\s)[Фф](\.|-)?1(\-)?(у|У)?($|\s)"
            }
            requirement = RegexpInTemplateRequirement(items)
            params = {"payload": {
                "userInfo": {
                    "tbcode": "32",
                },
                "message": "карточки конг фу 1"
            }}
            user = PicklableMock()
            user.parametrizer = PicklableMock()
            user.parametrizer.collect = Mock(return_value=params)
            self.assertFalse(requirement.check(None, user))



# coding: utf-8
from datetime import datetime

from pythonjsonlogger import jsonlogger
import logging
from core.model.factory import build_factory
from core.model.registered import Registered

loggers_formatter = Registered()

loggers_formatter_factory = build_factory(loggers_formatter)


class SmartKitJsonFormatter(jsonlogger.JsonFormatter):
    VERSION = 0
    DEV_TEAM = "NA"
    APPLICATION_NAME = "NA"

    def add_fields(self, log_record, record, message_dict):
        super(SmartKitJsonFormatter, self).add_fields(log_record, record, message_dict)
        dt = datetime.fromtimestamp(record.created)
        st = dt.strftime("%Y-%m-%dT%H:%M:%S")
        log_record['timestamp'] = "%s.%06d" % (st, record.msecs * 1000)
        log_record['version'] = self.VERSION
        log_record['team'] = self.DEV_TEAM
        log_record['application'] = self.APPLICATION_NAME
        if isinstance(record.args, dict):
            log_record['args'] = record.args

    def format(self, record: logging.LogRecord) -> str:
        # В готовый json логов добавляем поле с размером json-а.
        result = super(SmartKitJsonFormatter, self).format(record)
        log_size = len(result)
        # Убираем последнюю закрывающую скобку и добавляем поле log_size и закрываем скобки
        return f'{result[:-1]},"log_size":{log_size}}}'



from functools import cached_property

from core.message.from_message import SmartAppFromMessage
from smart_kit.message.smartapp_to_message import SmartAppToMessage


class AsIsToMessage(SmartAppToMessage):

    @cached_property
    def as_dict(self):
        self.incoming_message: SmartAppFromMessage
        fields = self.command.raw
        fields.update(self.root_nodes)
        return fields



from typing import Dict

from confluent_kafka.cimpl import Message as KafkaMessage
from core.mq.kafka.kafka_publisher import KafkaPublisher
from core.request.base_request import BaseRequest
from core.logging.logger_utils import log
import core.logging.logger_constants as log_const


class KafkaRequest(BaseRequest):
    TOPIC_KEY = "topic_key"
    KAFKA_KEY = "kafka_key"
    TOPIC = "topic"

    def __init__(self, items, id=None):
        super(KafkaRequest, self).__init__(items)
        items = items or {}
        self.topic_key = items.get(self.TOPIC_KEY)
        self.kafka_key = items.get(self.KAFKA_KEY)
        # topic_key has priority over topic
        self.topic = items.get(self.TOPIC)

    def update_empty_items(self, items: Dict[str, str]):
        self.topic_key = self.topic_key or items.get(self.TOPIC_KEY)
        self.kafka_key = self.kafka_key or items.get(self.KAFKA_KEY)
        self.topic = self.topic or items.get(self.TOPIC)

    @property
    def group_key(self):
        return "{}_{}".format(self.kafka_key, self.topic_key) if (self.topic_key and self.kafka_key) else None

    def _get_new_headers(self, source_mq_message: KafkaMessage):
        headers = source_mq_message.headers() or []
        return headers

    def send(self, data, publisher: KafkaPublisher, source_mq_message):
        headers = self._get_new_headers(source_mq_message)
        if self.topic_key is not None:
            publisher.send(data, source_mq_message.key(), self.topic_key, headers=headers)
        elif self.topic is not None:
            publisher.send_to_topic(data, source_mq_message.key(), self.topic, headers=headers)
        else:
            log_params = {
                "data": str(data),
                log_const.KEY_NAME: log_const.EXCEPTION_VALUE
            }
            log("KafkaRequest: got no topic and no topic_key", params=log_params, level="ERROR")

    def run(self, data, params):
        publishers = params["publishers"]
        publisher = publishers[self.kafka_key]
        self.send(data=data, publisher=publisher, source_mq_message=params["mq_message"])

    def __str__(self):
        if self.topic_key is not None:
            return f"KafkaRequest: topic_key={self.topic_key} kafka_key={self.kafka_key}"
        elif self.topic is not None:
            return f"KafkaRequest: topic={self.topic} kafka_key={self.kafka_key}"
        else:
            return f"KafkaRequest: kafka_key={self.kafka_key}"



# coding: utf-8
from typing import TypeVar, Union

T = TypeVar("T")


class Registered(dict):

    def __getitem__(self, key: Union[str, T]) -> T:
        value = self.get(key, key)
        assert not isinstance(value, str), "{} factory is not registered".format(key)
        return value


registered_factories = Registered()



# coding: utf-8
import dill

import core.logging.logger_constants as log_const
from core.logging.logger_utils import log
from core.repositories.base_repository import BaseRepository


class DillRepository(BaseRepository):
    """
    load unpickleable file
    Args:
        filename: path to the file
        source: file adapter e.g OSAdapter, CephAdapter etc
        required: raise FileNotFoundError if file not found
    """
    def __init__(self, filename, source=None, required=True, *args, **kwargs):
        super(DillRepository, self).__init__(source=source, *args, **kwargs)
        self.filename = filename
        self.required = required

    def load(self):
        dill._dill._reverse_typemap['ClassType'] = type
        try:

            with self.source.open(self.filename, 'rb') as stream:
                data = dill.load(stream)
                self.fill(data)
        except FileNotFoundError as error:
            params = {
                'error': str(error),
                log_const.KEY_NAME: log_const.EXCEPTION_VALUE
            }
            log('DillRepository.load loading failed. Error %(error)s', params=params, level='WARNING')
            if self.required:
                raise
        super(DillRepository, self).load()



import os
import time

from core.db_adapter.os_adapter import OSAdapter

import core.logging.logger_constants as log_const
from core.repositories.items_repository import ItemsRepository
from core.logging.logger_utils import log


class FileRepository(ItemsRepository):
    def __init__(self, filename, loader, source=None, save_target=None, saver=None, *args, **kwargs):
        super(FileRepository, self).__init__(source=source, *args, **kwargs)
        self.filename = filename
        self.loader = loader
        self.saver = saver
        self.save_target = save_target
        self._file_exist = False

    def load(self):
        if self.source.path_exists(self.filename):
            self._file_exist = True
            with self.source.open(self.filename, 'rb') as stream:
                binary_data = stream.read()
                data = binary_data.decode()
                self.fill(self.loader(data))
        else:
            self._file_exist = False
            params = {
                "error_repository_path": self.filename,
                log_const.KEY_NAME: log_const.EXCEPTION_VALUE
            }
            log("FileRepository.load loading failed with file %(error_repository_path)s",
                params=params, level="WARNING")
        super(FileRepository, self).load()

    def save(self, save_parameters):
        with self.source.open(self.save_target, 'wb') as stream:
            stream.write(self.saver(self.data, **save_parameters).encode())


class UpdatableFileRepository(FileRepository):
    def __init__(self, *args, update_cooldown=5, **kwargs):
        super().__init__(*args, **kwargs)
        self._last_mtime = None
        self._last_update_time = 0
        self.update_cooldown = update_cooldown
        if self.source and not isinstance(self.source, OSAdapter):
            raise Exception(f"{self.__class__.__name__} support only OSAdapter")

    @FileRepository.data.getter
    def data(self):
        if self._is_outdated:
            params = {
                "filename": self.filename
            }
            log("FileRepository.data %(filename)s is outdated. Data will be reloaded.",
                params=params, level="INFO")

            self.load()
        return self._data

    def load(self):
        super().load()
        if self._file_exist:
            self._last_mtime = self.source.mtime(self.filename)
        self._last_update_time = time.time()

    @property
    def expired(self):
        return self._last_update_time + self.update_cooldown < time.time()

    @property
    def _is_outdated(self):
        if self._file_exist and self.expired:
            return self.source.mtime(self.filename) > self._last_mtime
        return False



# coding: utf-8
from unittest import TestCase
from unittest.mock import Mock

from scenarios.scenario_models.field.composite_field import CompositeField
from scenarios.scenario_models.field.field import field_models
from smart_kit.utils.picklable_mock import PicklableMock


class MockDescr1():
    def __init__(self, *args, **kwargs):
        self.id = "d1"


class MockDescr2():
    def __init__(self, *args, **kwargs):
        self.id = "d2"


class MockItem1():
    def __init__(self, *args, **kwargs):
        self.available = False
        self.value = "raw"
        pass

    def set_available(self):
        pass

    def fill(self, value):
        self.value = value

    @property
    def valid(self):
        return True

    @property
    def raw(self):
        return self.value


class MockItem2(MockItem1):
    pass


field_models[MockDescr1] = MockItem1
field_models[MockDescr2] = MockItem2


class TestCompositeField(TestCase):
    def setUp(self):
        descr1 = MockDescr1(id="descr1")
        descr2 = MockDescr2(id="descr2")
        descriptions = {"d1": descr1, "d2": descr2}
        user = PicklableMock()
        fields = {"fields": {"d1": {}, "d2": {}}}
        self.composite_field = CompositeField(Mock(fields=descriptions), fields, user, None)

    def test_available1(self):
        self.assertFalse(self.composite_field.available)

    def test_available2(self):
        self.composite_field.fields["d1"].available = True
        self.composite_field.fields["d2"].available = True
        self.assertTrue(self.composite_field.available)

    def test_set_available(self):
        self.composite_field.set_available()
        for field_descr in self.composite_field.fields:
            self.composite_field.fields[field_descr].available = True

    def test_raw(self):
        expected = {'fields': {'d1': 'raw', 'd2': 'raw'}}
        raw = self.composite_field.raw
        self.assertEqual(expected, raw)

    def test_fill(self):
        expected = {'fields': {'d1': 'v1', 'd2': 'v2'}}
        value = {"d1": "v1", "d2": "v2"}
        self.composite_field.fill(value)
        raw = self.composite_field.raw
        self.assertEqual(expected, raw)

    def test_value(self):
        expected = {'d1': 'raw', 'd2': 'raw'}
        result = self.composite_field.value
        self.assertEqual(expected, result)

    def test_valid(self):
        result = self.composite_field.valid
        self.assertTrue(result)



import sys

from smart_kit.management.base import HelpCommand

from smart_kit.management.smart_kit_manager import SmartKitManager, CreateAppCommand, VersionCommand

if __name__ == "__main__":
    Manager = SmartKitManager()
    Manager.register_command("create_app", CreateAppCommand)
    Manager.register_command("version", VersionCommand)
    Manager.register_command("help", HelpCommand, Manager.commands)
    Manager.execute_from_command_line(sys.argv)



import unittest
from typing import Dict, Any, Union, Optional
from unittest.mock import Mock, ANY

from core.basic_models.actions.basic_actions import Action, action_factory, actions
from core.basic_models.actions.variable_actions import SetVariableAction, DeleteVariableAction, ClearVariablesAction
from core.model.registered import registered_factories
from scenarios.actions.action import (
    ChoiceScenarioAction,
    ClearCurrentScenarioAction,
    ClearCurrentScenarioFormAction,
    ClearScenarioByIdAction,
    CompositeFillFieldAction,
    FillFieldAction,
    SelfServiceActionWithState,
    SaveBehaviorAction,
    ResetCurrentNodeAction,
    RunScenarioAction,
    RunLastScenarioAction,
    AddHistoryEventAction
)
from scenarios.actions.action import ClearFormAction, ClearInnerFormAction, BreakScenarioAction, \
    RemoveFormFieldAction, RemoveCompositeFormFieldAction
from scenarios.scenario_models.history import Event
from smart_kit.action.smart_geo_action import SmartGeoAction
from smart_kit.message.smartapp_to_message import SmartAppToMessage
from smart_kit.utils.picklable_mock import PicklableMock, PicklableMagicMock


class MockAction:
    def __init__(self, items=None):
        self.items = items or {}
        self.result = items.get("result")

    def run(self, user, text_preprocessing_result, params=None):
        return self.result or ["test action run"]


class MockParametrizer:
    def __init__(self, user, items=None):
        self.items = items or {}
        self.user = user
        self.data = items.get("data") or {}
        self.filter = items.get("filter") or False

    def collect(self, text_preprocessing_result=None, filter_params=None):
        data = {
            "person_info": self.user.person_info.raw,
            "payload": self.user.message.payload}
        data.update(self.data)
        if self.filter:
            data.update({"filter": "filter_out"})
        return data


class ClearFormIdActionTest(unittest.TestCase):
    def test_run(self):
        action = ClearFormAction({"form": "form"})
        user = PicklableMagicMock()
        action.run(user, None)
        user.forms.remove_item.assert_called_once_with("form")


class ClearInnerFormActionTest(unittest.TestCase):
    def test_run(self):
        action = ClearInnerFormAction({"form": "form", "inner_form": "inner_form"})
        user, form = PicklableMagicMock(), PicklableMagicMock()
        user.forms.__getitem__.return_value = form
        action.run(user, None)
        form.forms.remove_item.assert_called_once_with("inner_form")


class BreakScenarioTest(unittest.TestCase):
    def test_run_1(self):
        scenario_id = "test_id"
        action = BreakScenarioAction({"scenario_id": scenario_id})
        user = PicklableMock()
        scenario_model = PicklableMagicMock()
        scenario_model.set_break = Mock(return_value=None)
        user.scenario_models = {scenario_id: scenario_model}
        action.run(user, None)
        user.scenario_models[scenario_id].set_break.assert_called_once()

    def test_run_2(self):
        scenario_id = "test_id"
        action = BreakScenarioAction({})
        user = PicklableMock()
        user.last_scenarios.last_scenario_name = "test_id"
        scenario_model = PicklableMagicMock()
        scenario_model.set_break = Mock(return_value=None)
        user.scenario_models = {scenario_id: scenario_model}
        action.run(user, None)
        user.scenario_models[scenario_id].set_break.assert_called_once()


class RemoveFormFieldActionTest(unittest.TestCase):
    def test_run(self):
        action = RemoveFormFieldAction({"form": "form", "field": "field"})
        user, form = PicklableMagicMock(), PicklableMagicMock()
        user.forms.__getitem__.return_value = form
        action.run(user, None)
        form.fields.remove_item.assert_called_once_with("field")


class RemoveCompositeFormFieldActionTest(unittest.TestCase):
    def test_run(self):
        action = RemoveCompositeFormFieldAction({"form": "form", "inner_form": "form", "field": "field"})
        user, inner_form, form = PicklableMagicMock(), PicklableMagicMock(), PicklableMagicMock()
        form.forms.__getitem__.return_value = inner_form
        user.forms.__getitem__.return_value = form
        action.run(user, None)
        inner_form.fields.remove_item.assert_called_once_with("field")


class SaveBehaviorActionTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        user = PicklableMock()
        user.message = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        user.last_scenarios.last_scenario_name = "scenario_id"
        test_incremental_id = "test_incremental_id"
        user.message.incremental_id = test_incremental_id
        cls.user = user

    def test_save_behavior_scenario_name(self):
        data = {"behavior": "test"}
        behavior = PicklableMock()
        behavior.add = PicklableMock()
        self.user.behaviors = behavior
        action = SaveBehaviorAction(data)
        tpr = PicklableMock()
        tpr_raw = tpr.raw
        action.run(self.user, tpr)
        self.user.behaviors.add.assert_called_once_with(self.user.message.generate_new_callback_id(), "test",
                                                        self.user.last_scenarios.last_scenario_name, tpr_raw,
                                                        action_params=None)

    def test_save_behavior_without_scenario_name(self):
        data = {"behavior": "test", "check_scenario": False}
        behavior = PicklableMock()
        behavior.add = PicklableMock()
        self.user.behaviors = behavior
        action = SaveBehaviorAction(data)
        text_preprocessing_result_raw = PicklableMock()
        text_preprocessing_result = Mock(raw=text_preprocessing_result_raw)
        action.run(self.user, text_preprocessing_result, None)
        self.user.behaviors.add.assert_called_once_with(self.user.message.generate_new_callback_id(), "test", None,
                                                        text_preprocessing_result_raw, action_params=None)


class SelfServiceActionWithStateTest(unittest.TestCase):
    def setUp(self) -> None:
        self.user = PicklableMock()
        self.user.settings = {"template_settings": {"self_service_with_state_save_messages": True}}

    def test_action_1(self):
        data = {"behavior": "test", "check_scenario": False, "command_action": {"command": "cmd_id", "nodes": {},
                                                                                "request_data": {}}}
        registered_factories[Action] = action_factory
        actions["action_mock"] = MockAction
        self.user.parametrizer = MockParametrizer(self.user, {})
        self.user.message = PicklableMock()
        local_vars = PicklableMock()
        local_vars.values = dict()
        self.user.local_vars = local_vars
        test_incremental_id = "test_incremental_id"
        self.user.message.incremental_id = test_incremental_id
        behavior = PicklableMock()
        behavior.check_got_saved_id = Mock(return_value=False)
        self.user.behaviors = behavior
        action = SelfServiceActionWithState(data)
        text_preprocessing_result_raw = PicklableMock()
        text_preprocessing_result = Mock(raw=text_preprocessing_result_raw)
        result = action.run(self.user, text_preprocessing_result, None)
        behavior.check_got_saved_id.assert_called_once()
        behavior.add.assert_called_once()
        self.assertEqual(result[0].name, "cmd_id")
        self.assertEqual(result[0].raw, {'messageName': 'cmd_id', 'payload': {}})

    def test_action_2(self):
        data = {"behavior": "test", "check_scenario": False, "command_action": {"command": "cmd_id", "nodes": {}}}
        self.user.parametrizer = MockParametrizer(self.user, {})
        self.user.message = PicklableMock()
        test_incremental_id = "test_incremental_id"
        self.user.message.incremental_id = test_incremental_id
        behavior = PicklableMock()
        self.user.behaviors = behavior
        behavior.check_got_saved_id = Mock(return_value=True)
        action = SelfServiceActionWithState(data)
        result = action.run(self.user, None)
        behavior.add.assert_not_called()
        self.assertIsNone(result)

    def test_action_3(self):
        data = {"behavior": "test", "command_action": {"command": "cmd_id", "nodes": {}, "request_data": {}}}
        registered_factories[Action] = action_factory
        actions["action_mock"] = MockAction
        self.user.parametrizer = MockParametrizer(self.user, {})
        self.user.message = PicklableMock()
        self.user.message = PicklableMock()
        local_vars = PicklableMock()
        local_vars.values = dict()
        self.user.local_vars = local_vars
        test_incremental_id = "test_incremental_id"
        self.user.message.incremental_id = test_incremental_id
        _new_behavior_id = PicklableMock()
        self.user.message.generate_new_callback_id = lambda: _new_behavior_id
        behavior = PicklableMock()
        behavior.check_got_saved_id = Mock(return_value=False)
        behavior.add = PicklableMock()
        self.user.behaviors = behavior
        self.user.last_scenarios = PicklableMock()
        scenarios_names = ["test_scenario"]
        self.user.last_scenarios.last_scenario_name = "test_scenario"
        self.user.last_scenarios.scenarios_names = scenarios_names
        action = SelfServiceActionWithState(data)
        text_preprocessing_result_raw = PicklableMock()
        text_preprocessing_result = Mock(raw=text_preprocessing_result_raw)
        result = action.run(self.user, text_preprocessing_result, None)
        behavior.check_got_saved_id.assert_called_once()
        behavior.add.assert_called_once()
        self.assertEqual(result[0].name, "cmd_id")
        self.assertEqual(result[0].raw, {'messageName': 'cmd_id', 'payload': {}})
        behavior.add.assert_called_once_with(
            self.user.message.generate_new_callback_id(), "test", scenarios_names[-1], text_preprocessing_result_raw,
            ANY
        )


class SetVariableActionTest(unittest.TestCase):

    def setUp(self):
        template = PicklableMock()
        template.get_template = Mock(return_value=[])
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        user.message = PicklableMock()
        user.person_info = PicklableMock()
        user.descriptions = {"render_templates": template}
        user.variables = PicklableMagicMock()
        user.variables.values = {}
        user.variables.set = PicklableMock()
        self.user = user

    def test_action(self):
        action = SetVariableAction({"key": "some_key", "value": "some_value"})
        action.run(self.user, None)
        self.user.variables.set.assert_called_with("some_key", "some_value", None)

    def test_action_jinja_key_default(self):
        self.user.message.payload = {"some_value": "some_value_test"}
        action = SetVariableAction({"key": "some_key", "value": "{{payload.some_value}}"})
        action.run(self.user, None)
        self.user.variables.set.assert_called_with("some_key", "some_value_test", None)

    def test_action_jinja_no_key(self):
        self.user.message.payload = {"some_value": "some_value_test"}
        action = SetVariableAction({"key": "some_key", "value": "{{payload.no_key}}"})
        action.run(self.user, None)
        self.user.variables.set.assert_called_with("some_key", "", None)


class DeleteVariableActionTest(unittest.TestCase):

    def setUp(self):
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        user.message = PicklableMock()
        user.message.payload = {"some_value": "some_value_test"}
        user.person_info = PicklableMock()
        user.variables = PicklableMagicMock()
        user.variables.delete = PicklableMock()
        self.user = user

    def test_action(self):
        action = DeleteVariableAction({"key": "some_key_1"})
        action.run(self.user, None)
        self.user.variables.delete.assert_called_with("some_key_1")


class ClearVariablesActionTest(unittest.TestCase):

    def setUp(self):
        self.var_value = {
            "some_key_1": "some_value_1",
            "some_key_2": "some_value_2",
        }
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        user.message = PicklableMock()
        user.person_info = PicklableMock()
        user.variables = PicklableMagicMock()
        user.variables.values = self.var_value
        user.variables.clear = PicklableMock()
        self.user = user

    def test_action(self):
        action = ClearVariablesAction()
        action.run(self.user, None)
        self.user.variables.clear.assert_called_with()


class FillFieldActionTest(unittest.TestCase):

    def test_fill_field(self):
        params = {"test_field": "test_data"}
        data = {"form": "test_form", "field": "test_field", "data_path": "{{test_field}}"}
        action = FillFieldAction(data)
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {"data": params})
        user.forms = {"test_form": PicklableMock()}
        field = PicklableMock()
        field.fill = PicklableMock()
        user.forms["test_form"].fields = {"test_field": field}
        action.run(user, None)
        field.fill.assert_called_once_with(params["test_field"])


class CompositeFillFieldActionTest(unittest.TestCase):

    def test_fill_field(self):
        params = {"test_field": "test_data"}
        data = {"form": "test_form", "field": "test_field", "internal_form": "test_internal_form",
                "data_path": "{{test_field}}", "parametrizer": {"data": params}}
        action = CompositeFillFieldAction(data)
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {"data": params})
        form = PicklableMock()
        internal_form = PicklableMock()
        form.forms = {"test_internal_form": internal_form}
        user.forms = {"test_form": form}
        field = PicklableMock()
        field.fill = PicklableMock()
        user.forms["test_form"].forms["test_internal_form"].fields = {"test_field": field}
        action.run(user, None)
        field.fill.assert_called_once_with(params["test_field"])


class ScenarioActionTest(unittest.TestCase):
    def test_scenario_action(self):
        action = RunScenarioAction({"scenario": "test"})
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        scen = PicklableMock()
        scen_result = ['done']
        scen.run.return_value = scen_result
        user.descriptions = {"scenarios": {"test": scen}}
        result = action.run(user, PicklableMock())
        self.assertEqual(result, scen_result)

    def test_scenario_action_with_jinja_good(self):
        params = {'next_scenario': 'ANNA.pipeline.scenario'}
        items = {"scenario": "{{next_scenario}}"}

        action = RunScenarioAction(items)
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {"data": params})
        scen = PicklableMock()
        scen_result = ['done']
        scen.run.return_value = scen_result
        user.descriptions = {"scenarios": {"ANNA.pipeline.scenario": scen}}
        result = action.run(user, PicklableMock())
        self.assertEqual(result, scen_result)

    def test_scenario_action_no_scenario(self):
        action = RunScenarioAction({"scenario": "{{next_scenario}}"})
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        scen = PicklableMock()
        scen_result = 'done'
        scen.run.return_value = scen_result
        user.descriptions = {"scenarios": {"next_scenario": scen}}
        result = action.run(user, PicklableMock())
        self.assertEqual(result, [])

    def test_scenario_action_without_jinja(self):
        action = RunScenarioAction({"scenario": "next_scenario"})
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        scen = PicklableMock()
        scen_result = ['done']
        scen.run.return_value = scen_result
        user.descriptions = {"scenarios": {"next_scenario": scen}}
        result = action.run(user, PicklableMock())
        self.assertEqual(result, scen_result)


class RunLastScenarioActionTest(unittest.TestCase):

    def test_scenario_action(self):
        action = RunLastScenarioAction({})
        user = PicklableMock()
        scen = PicklableMock()
        scen_result = ['done']
        scen.run.return_value = scen_result
        user.descriptions = {"scenarios": {"test": scen}}
        user.last_scenarios = PicklableMock()
        last_scenario_name = "test"
        user.last_scenarios.scenarios_names = [last_scenario_name]
        user.last_scenarios.last_scenario_name = last_scenario_name
        result = action.run(user, PicklableMock())
        self.assertEqual(result, scen_result)


class ChoiceScenarioActionTest(unittest.TestCase):

    @staticmethod
    def mock_and_perform_action(test_items: Dict[str, Any], expected_result: Optional[str] = None,
                                expected_scen: Optional[str] = None) -> Union[str, None]:
        action = ChoiceScenarioAction(test_items)
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        scen = PicklableMock()
        scen.run.return_value = expected_result
        if expected_scen:
            user.descriptions = {"scenarios": {expected_scen: scen}}
        return action.run(user, PicklableMock())

    def test_choice_scenario_action(self):
        # Проверяем, что запустили нужный сценарий, в случае если выполнился его requirement
        test_items = {
            "scenarios": [
                {
                    "scenario": "test_1",
                    "requirement": {"cond": False}
                },
                {
                    "scenario": "test_2",
                    "requirement": {"cond": False}
                },
                {
                    "scenario": "test_N",
                    "requirement": {"cond": True}
                }
            ],
            "else_action": {"type": "test", "result": "ELSE ACTION IS DONE"}
        }
        expected_scen_result = ["test_N_done"]
        real_scen_result = self.mock_and_perform_action(
            test_items, expected_result=expected_scen_result, expected_scen="test_N")
        self.assertEqual(real_scen_result, expected_scen_result)

    def test_choice_scenario_action_no_else_action(self):
        # Проверяем, что вернули None в случае если ни один сценарий не запустился (requirement=False) и else_action нет
        test_items = {
            "scenarios": [
                {
                    "scenario": "test_1",
                    "requirement": {"cond": False}
                },
                {
                    "scenario": "test_2",
                    "requirement": {"cond": False}
                }
            ]
        }
        real_scen_result = self.mock_and_perform_action(test_items)
        self.assertEqual([], real_scen_result)

    def test_choice_scenario_action_with_else_action(self):
        # Проверяем, что выполняется else_action в случае если ни один сценарий не запустился т.к их requirement=False
        test_items = {
            "scenarios": [
                {
                    "scenario": "test_1",
                    "requirement": {"cond": False}
                },
                {
                    "scenario": "test_2",
                    "requirement": {"cond": False}
                }
            ],
            "else_action": {"type": "test", "result": ["ELSE ACTION IS DONE"]}
        }
        expected_scen_result = ["ELSE ACTION IS DONE"]
        real_scen_result = self.mock_and_perform_action(test_items, expected_result=expected_scen_result)
        self.assertEqual(real_scen_result, expected_scen_result)


class ClearCurrentScenarioActionTest(unittest.TestCase):

    def test_action(self):
        scenario_name = "test_scenario"
        user = PicklableMock()
        user.forms.remove_item = PicklableMock()

        user.last_scenarios.last_scenario_name = scenario_name
        user.last_scenarios.delete = PicklableMock()
        scenario = PicklableMock()
        scenario.form_type = scenario_name
        user.descriptions = {"scenarios": {scenario_name: scenario}}

        action = ClearCurrentScenarioAction({})
        result = action.run(user, {}, {})
        self.assertEqual([], result)
        user.last_scenarios.delete.assert_called_once()
        user.forms.remove_item.assert_called_once()

    def test_action_with_empty_scenarios_names(self):
        user = PicklableMock()
        user.forms.remove_item = PicklableMock()

        user.last_scenarios.last_scenario_name = None
        user.last_scenarios.delete = PicklableMock()

        action = ClearCurrentScenarioAction({})
        result = action.run(user, {}, {})
        self.assertEqual([], result)
        user.last_scenarios.delete.assert_not_called()
        user.forms.remove_item.assert_not_called()


class ClearScenarioByIdActionTest(unittest.TestCase):

    def test_action(self):
        scenario_name = "test_scenario"
        user = PicklableMock()
        user.forms = PicklableMock()

        user.last_scenarios.last_scenario_name = scenario_name
        scenario = PicklableMock()
        scenario.form_type = scenario_name
        user.descriptions = {"scenarios": {scenario_name: scenario}}

        action = ClearScenarioByIdAction({"scenario_id": scenario_name})
        result = action.run(user, {}, {})
        self.assertEqual([], result)
        user.last_scenarios.delete.assert_called_once()
        user.forms.remove_item.assert_called_once()

    def test_action_with_empty_scenarios_names(self):
        user = PicklableMock()
        user.forms = PicklableMock()

        user.last_scenarios.last_scenario_name = "test_scenario"

        action = ClearScenarioByIdAction({})
        result = action.run(user, {}, {})
        self.assertEqual([], result)
        user.last_scenarios.delete.assert_not_called()
        user.forms.remove_item.assert_not_called()


class ClearCurrentScenarioFormActionTest(unittest.TestCase):
    def test_action(self):
        scenario_name = "test_scenario"
        user = PicklableMock()
        user.forms = PicklableMock()
        user.forms.remove_item = PicklableMock()

        user.last_scenarios.last_scenario_name = scenario_name
        scenario = PicklableMock()
        scenario.form_type = scenario_name
        scenario.keep_forms_alive = False
        user.descriptions = {"scenarios": {scenario_name: scenario}}

        action = ClearCurrentScenarioFormAction({})
        result = action.run(user, {}, {})
        self.assertEqual([], result)
        user.forms.clear_form.assert_called_once()

    def test_action_with_empty_last_scenario(self):
        scenario_name = "test_scenario"
        user = PicklableMock()
        user.forms = PicklableMock()
        user.forms.remove_item = PicklableMock()

        user.last_scenarios.last_scenario_name = None
        scenario = PicklableMock()
        scenario.form_type = scenario_name
        scenario.keep_forms_alive = False
        user.descriptions = {"scenarios": {scenario_name: scenario}}

        action = ClearCurrentScenarioFormAction({})
        result = action.run(user, {}, {})
        self.assertEqual([], result)
        user.forms.remove_item.assert_not_called()


class ResetCurrentNodeActionTest(unittest.TestCase):
    def test_action(self):
        user = PicklableMock()
        user.forms = PicklableMock()
        user.last_scenarios.last_scenario_name = 'test_scenario'
        scenario_model = PicklableMock()
        scenario_model.current_node = 'some_node'
        user.scenario_models = {'test_scenario': scenario_model}

        action = ResetCurrentNodeAction({})
        result = action.run(user, {}, {})
        self.assertEqual([], result)
        self.assertIsNone(user.scenario_models['test_scenario'].current_node)

    def test_action_with_empty_last_scenario(self):
        user = PicklableMock()
        user.forms = PicklableMock()
        user.last_scenarios.last_scenario_name = None
        scenario_model = PicklableMock()
        scenario_model.current_node = 'some_node'
        user.scenario_models = {'test_scenario': scenario_model}

        action = ResetCurrentNodeAction({})
        result = action.run(user, {}, {})
        self.assertEqual([], result)
        self.assertEqual('some_node', user.scenario_models['test_scenario'].current_node)

    def test_specific_target(self):
        user = PicklableMock()
        user.forms = PicklableMock()
        user.last_scenarios.last_scenario_name = 'test_scenario'
        scenario_model = PicklableMock()
        scenario_model.current_node = 'some_node'
        user.scenario_models = {'test_scenario': scenario_model}

        items = {
            'node_id': 'another_node'
        }
        action = ResetCurrentNodeAction(items)
        result = action.run(user, {}, {})
        self.assertEqual([], result)
        self.assertEqual('another_node', user.scenario_models['test_scenario'].current_node)


class AddHistoryEventActionTest(unittest.TestCase):

    def setUp(self):
        main_form = PicklableMock()
        main_form.field_1 = "value_1"
        parametrizer = PicklableMock()
        message = PicklableMock()
        message.name = "CLIENT_INFO_RESPONSE"
        parametrizer.collect = Mock(return_value={"message": message, "main_form": main_form})

        self.user = Mock(parametrizer=parametrizer)
        self.user.history = PicklableMock()
        self.user.history.add_event = PicklableMock()
        self.user.last_scenarios.last_scenario_name = 'test_scenario'

    def test_action_with_non_empty_scenario(self):
        scenario = PicklableMock()
        scenario.id = 'name'
        scenario.version = '1.0'
        self.user.descriptions = {'scenarios': {'test_scenario': scenario}}
        items = {
            'event_type': 'type',
            'event_content': {'foo': 'bar'},
            'results': 'result'
        }
        expected = Event(
            type='type',
            result='result',
            content={'foo': 'bar'},
            scenario='name',
            scenario_version='1.0'
        )

        action = AddHistoryEventAction(items)
        action.run(self.user, None, None)

        self.user.history.add_event.assert_called_once()
        self.user.history.add_event.assert_called_once_with(expected)

    def test_action_with_empty_scenario(self):
        self.user.descriptions = {'scenarios': {}}
        items = {
            'event_type': 'type',
            'event_content': {'foo': 'bar'},
            'results': 'result'
        }

        action = AddHistoryEventAction(items)
        action.run(self.user, None, None)

        self.user.history.add_event.assert_not_called()

    def test_action_with_jinja(self):
        scenario = PicklableMock()
        scenario.id = 'name'
        scenario.version = '1.0'
        self.user.descriptions = {'scenarios': {'test_scenario': scenario}}
        items = {
            'event_type': 'type',
            'event_content': {'field_1': '{{ main_form.field_1 }}'},
            'results': '{{ message.name }}'
        }
        expected = Event(
            type='type',
            result='CLIENT_INFO_RESPONSE',
            content={'field_1': 'value_1'},
            scenario='name',
            scenario_version='1.0'
        )

        action = AddHistoryEventAction(items)
        action.run(self.user, None, None)

        self.user.history.add_event.assert_called_once()
        self.user.history.add_event.assert_called_once_with(expected)


class SmartGeoActionTest(unittest.TestCase):

    def setUp(self):
        items = {"type": "smart_geo"}
        self.smart_geo_action = SmartGeoAction(items)

    def test_action_send_request(self):
        incoming_message = Mock(incremental_id="1605196199186625000",
                                session_id="0062530b-5521-42cc-90b0-a9d65dea4e98",
                                uuid={"userChannel": "B2C", "userId": "ec8a9097-1508-4bec-8d97-67f2329c03e0",
                                      "sub": "385342565001000018390f1f"},
                                payload={})
        user = Mock()
        text_preprocessing_result = Mock()
        params = Mock()
        command = self.smart_geo_action.run(user, text_preprocessing_result, params)[0]
        answer = SmartAppToMessage(command, incoming_message, None)
        expected = {
            "messageId": "1605196199186625000",
            "sessionId": "0062530b-5521-42cc-90b0-a9d65dea4e98",
            "messageName": "GET_PROFILE_DATA",
            "uuid": {
                "userId": "ec8a9097-1508-4bec-8d97-67f2329c03e0",
                "userChannel": "B2C",
                "sub": "385342565001000018390f1f"
            },
            "payload": {
            }
        }
        self.assertEqual(answer.as_dict, expected)



import os
from core.repositories.base_repository import BaseRepository


class ShardRepository(BaseRepository):

    def __init__(self, path, loader, source=None, *args, **kwargs):
        super(ShardRepository, self).__init__(source=source, *args, **kwargs)
        self.path = path
        self.loader = loader

    @staticmethod
    def join_path(path, file_name):
        path = os.path.join(path, file_name)
        path = os.path.normpath(path)
        return path

    @staticmethod
    def _get_data_type(data):
        t = None
        for k, v in data.items():
            if not t:
                if isinstance(v, dict):
                    t = dict
                elif isinstance(v, list):
                    t = list
            if t and not isinstance(v, t):
                raise TypeError(
                    f"There are more than one data type in folder repository: values {v} incompatible with"
                    f" data type {t}")
        return t

    def load(self):
        super(ShardRepository, self).load()

    def fill(self, data):
        res = None
        t = self._get_data_type(data)
        if t == dict:
            res = {}
            for k, v in data.items():
                res.update(v)
        elif t == list:
            res = []
            for k, v in data.items():
                res.extend(v)
        self.data = res

    def fill_on_top(self, data):
        t = self._get_data_type(data)
        for k, v in data.items():
            if t == dict:
                self.data.update(v)
            elif t == list:
                self.data.extend(v)

    def save(self, save_parameters):
        raise NotImplementedError



# coding: utf-8
from scenarios.user.message_history.message_history import MessageHistory


class PreprocessingScenariosMessages(MessageHistory):

    def add(self, text_preprocessing_result):
        if text_preprocessing_result is not None and text_preprocessing_result.original_text:
            self._push(text_preprocessing_result.raw, self.INCOMING)

    @property
    def processed_items(self):
        return [item['message'] for item in reversed(self.items)]



# coding: utf-8
import unittest

from core.basic_models.operators.comparators import MoreComparator, \
    LessComparator, MoreOrEqualComparator, LessOrEqualComparator, EqualComparator


class ComparatorsTest(unittest.TestCase):
    def test_more_compare_sucsess(self):
        comparator = MoreComparator({})
        assert comparator.compare(3, 1)

    def test_more_compare_fail(self):
        comparator = MoreComparator({})
        assert not comparator.compare(3, 10)

    def test_less_compare_sucsess(self):
        comparator = LessComparator({})
        assert comparator.compare(3, 10)

    def test_less_compare_fail(self):
        comparator = LessComparator({})
        assert not comparator.compare(3, 1)

    def test_more_or_equal_sucsess(self):
        comparator = MoreOrEqualComparator({})
        assert comparator.compare(1, 1)

    def test_more_or_equal_fail(self):
        comparator = MoreOrEqualComparator({})
        assert not comparator.compare(1, 3)

    def test_less_or_equal_sucsess(self):
        comparator = LessOrEqualComparator({})
        assert comparator.compare(1, 1)

    def test_less_or_equal_fail(self):
        comparator = LessOrEqualComparator({})
        assert not comparator.compare(3, 1)

    def test_equal_sucsess(self):
        comparator = EqualComparator({})
        assert comparator.compare(1, 1)

    def test_equal_fail(self):
        comparator = EqualComparator({})
        assert not comparator.compare(1, 2)



# coding: utf-8
import unittest
from unittest.mock import Mock

from smart_kit.handlers import handle_server_action


class HandlersTest3(unittest.TestCase):
    def setUp(self):
        self.app_name = "TestAppName"
        self.test_payload_1 = {"server_action": {}}
        self.test_payload_2 = {"server_action": {"action_id": 1, "parameters": 1}}
        self.test_user = Mock('user')

    def test_handle_server_action_init(self):
        obj = handle_server_action.HandlerServerAction(self.app_name)
        self.assertTrue(obj.get_action_params(self.test_payload_1) == {})
        self.assertIsNotNone(handle_server_action.SERVER_ACTION)

    def test_handle_server_action_get_action_name(self):
        obj = handle_server_action.HandlerServerAction(self.app_name)
        with self.assertRaises(KeyError):
            obj.get_action_name(self.test_payload_1, self.test_user)
        self.assertTrue(obj.get_action_name(self.test_payload_2, self.test_user) == 1)

    def test_handle_server_action_get_action_params(self):
        obj = handle_server_action.HandlerServerAction(self.app_name)
        self.assertTrue(obj.get_action_params(self.test_payload_1) == {})
        self.assertTrue(obj.get_action_params(self.test_payload_2) == 1)



import time

from core.model.registered import Registered
from scenarios.scenario_models.field.field import field_model_factory
from scenarios.scenario_models.field.fields import Fields

form_models = Registered()


def form_model_factory(value, description, user):
    _type = type(description)
    model = form_models[_type]
    return model(value, description, user)


class BaseForm:
    def __init__(self, items, description, user):
        items = items or {}
        self.description = description
        self._valid = self.description.valid
        self._user = user
        self.remove_time = items.get("remove_time")

    def touch(self):
        lifetime = self.description.lifetime
        if lifetime:
            self.remove_time = int(time.time()) + lifetime

    def check_expired(self):
        return time.time() >= self.remove_time if self.remove_time else False

    @property
    def raw(self):
        raise NotImplementedError

    def get_fields_values(self):
        raise NotImplementedError


class Form(BaseForm):
    def __init__(self, items, description, user):
        super(Form, self).__init__(items, description, user)
        items = items or {}
        self.fields = Fields(items.get("fields"), description.fields, user, field_model_factory, description.lifetime)

    def is_valid(self):
        if not self._valid:
            self._valid = all(self.fields[description].valid for description in self.fields)
        return self._valid

    def refresh(self):
        self.touch()
        for description in self.fields:
            if description.need_save_context:
                self._user.last_fields[description.id].set_remove_time(self.description.lifetime)

    def get_fields_values(self):
        return {self.description.id: self.fields.values}

    @property
    def raw(self):
        raw = {}
        if self.fields.raw:
            raw["fields"] = self.fields.raw
        if self.remove_time:
            raw["remove_time"] = self.remove_time
        return raw



from unittest import TestCase

from core.utils.stats_timer import StatsTimer


class TestStatsTimer(TestCase):

    def test_1(self):
        timer = StatsTimer()
        with timer:
            a = 3

        self.assertTrue(timer.msecs > 0)



class AppCommand:
    @classmethod
    def doc(cls):
        return cls.__doc__ or "No help provided"

    def execute(self, *args, **kwargs):
        raise NotImplementedError


class HelpCommand(AppCommand):
    def __init__(self, commands):
        self._commands = commands

    def execute(self, *args, **kwargs):
        if not args:
            print("Available commands:")
            for cmd in self._commands:
                print(f"\t {cmd}")
            return
        if len(args) >= 2:
            raise ValueError("Wrong command usage")

        name = args[0]
        if name not in self._commands:
            print(f"{name} command not registered")
            return

        command = self._commands[name]
        print(command.doc())


class Manager:
    def __init__(self):
        self.commands = {}

    def register_command(self, name, command, *args, **kwargs):
        self.commands[name] = command(*args, **kwargs)

    def execute_command(self, name, *args, **kwargs):
        return self.commands[name].execute(*args, **kwargs)

    def default(self, argv):
        return NotImplementedError

    def execute_from_command_line(self, argv):
        if len(argv) <= 1:
            return self.default(argv)

        name = argv[1]

        if name not in self.commands:
            return self.default(argv)

        return self.execute_command(name, *argv[2:])



# coding: utf-8
import functools
from collections import OrderedDict
from typing import TypeVar, Type, List, Callable, Any, Dict

from core.model.registered import registered_factories, Registered

T = TypeVar("T")


def build_factory(registry_models: Registered) -> Callable:
    def _inner(items, *args, **kwargs):
        type = items.get("type")
        model = registry_models[type]
        try:
            return model(items, *args, **kwargs)
        except:
            raise

    return _inner


def factory(type: Type[T], **params) -> Callable[[Type[T]], Callable]:
    def _inner(func: Callable) -> Callable:
        @functools.wraps(func)
        def _wrap(self, *args: Any, **kwargs: Any) -> T:
            result = func(self, *args, **kwargs)
            result = result or {}
            cls: Type[T] = registered_factories[type]
            if not isinstance(result, dict):
                result = cls(result)
            else:
                result = cls(result or {}, **params)
            return result

        return _wrap

    return _inner


def list_factory(type: Type[T], **params) -> Callable[[Type[T]], Callable]:
    def _inner(func: Callable) -> Callable:
        @functools.wraps(func)
        def _wrap(*args: Any, **kwargs: Any) -> List[T]:
            cls: Type[T] = registered_factories[type]
            res: List[T] = []
            for items in func(*args, **kwargs) or []:
                if not isinstance(items, dict):
                    res.append(cls(items))
                else:
                    res.append(cls(items or {}, **params))
            return res

        return _wrap

    return _inner


def dict_factory(type: Type[T], has_id=True) -> Callable[[Type[T]], Callable]:
    def _inner(func: Callable) -> Callable:
        @functools.wraps(func)
        def _wrap(*args: Any, **kwargs: Any) -> Dict[str, T]:
            cls: Type[T] = registered_factories[type]
            items_iterator = func(*args, **kwargs).items()
            if has_id:
                return {id: cls(id=id, items=items) for id, items in items_iterator or {}}
            else:
                return {id: cls(items) for id, items in items_iterator or {}}

        return _wrap

    return _inner


def ordered_dict_factory(type: Type[T], has_id=True) -> Callable[[Type[T]], Callable]:
    def _inner(func: Callable) -> Callable:
        @functools.wraps(func)
        def _wrap(*args: Any, **kwargs: Any) -> 'OrderedDict[str, T]':
            cls: Type[T] = registered_factories[type]
            items_iterator = func(*args, **kwargs).items()
            result = OrderedDict({})
            for id, items in items_iterator:
                result[id] = cls(id=id, items=items) if has_id else cls(items)
            return result

        return _wrap

    return _inner



import os
from unittest import TestCase
from unittest.mock import patch

import smart_kit
from scenarios.scenario_models.field.field_filler_description import (
    IntersectionFieldFiller,
    IntersectionOriginalTextFiller
)
from smart_kit.text_preprocessing.local_text_normalizer import LocalTextNormalizer
from smart_kit.utils.picklable_mock import PicklableMock


def patch_get_app_config(mock_get_app_config):
    result = PicklableMock()
    sk_path = os.path.dirname(smart_kit.__file__)
    result.STATIC_PATH = os.path.join(sk_path, 'template/static')
    mock_get_app_config.return_value = result
    result.NORMALIZER = LocalTextNormalizer()
    mock_get_app_config.return_value = result


class TestIntersectionFieldFiller(TestCase):

    @patch('smart_kit.configs.get_app_config')
    def test_1(self, mock_get_app_config):
        patch_get_app_config(mock_get_app_config)
        expected = 'лосось'
        items = {
            'cases': {
                'лосось': [
                    'хорошая рыба'
                ],
                'килька': [
                    'консервы'
                ]
            }
        }
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.tokenized_elements_list_pymorphy = [
            {'lemma': 'весь'},
            {'lemma': 'хороший'},
            {'lemma': 'и'},
            {'lemma': 'спасибо'},
            {'lemma': 'за'},
            {'lemma': 'рыба'},
        ]

        filler = IntersectionFieldFiller(items)
        result = filler.extract(text_preprocessing_result, None)

        self.assertEqual(expected, result)

    @patch('smart_kit.configs.get_app_config')
    def test_2(self, mock_get_app_config):
        patch_get_app_config(mock_get_app_config)
        items = {
            'strict': True,
            'cases': {
                'лосось': [
                    'хорошая рыба'
                ],
                'килька': [
                    'консервы'
                ]
            }
        }
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.tokenized_elements_list_pymorphy = [
            {'lemma': 'весь'},
            {'lemma': 'хороший'},
            {'lemma': 'и'},
            {'lemma': 'спасибо'},
            {'lemma': 'за'},
            {'lemma': 'рыба'},
        ]

        filler = IntersectionFieldFiller(items)
        result = filler.extract(text_preprocessing_result, None)

        self.assertIsNone(result)

    @patch('smart_kit.configs.get_app_config')
    def test_3(self, mock_get_app_config):
        patch_get_app_config(mock_get_app_config)
        expected = 'лосось'
        items = {
            'strict': True,
            'cases': {
                'лосось': [
                    'хорошая рыба'
                ],
                'килька': [
                    'консервы'
                ]
            }
        }
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.tokenized_elements_list_pymorphy = [
            {'lemma': 'хороший'},
            {'lemma': 'рыба'},
        ]

        filler = IntersectionFieldFiller(items)
        result = filler.extract(text_preprocessing_result, None)

        self.assertEqual(expected, result)

    @patch('smart_kit.configs.get_app_config')
    def test_4(self, mock_get_app_config):
        patch_get_app_config(mock_get_app_config)
        items = {}
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.tokenized_elements_list_pymorphy = []

        filler = IntersectionFieldFiller(items)
        result = filler.extract(text_preprocessing_result, None)

        self.assertIsNone(result)

    @patch('smart_kit.configs.get_app_config')
    def test_5(self, mock_get_app_config):
        patch_get_app_config(mock_get_app_config)
        expected = 'дефолтный тунец'
        items = {
            'cases': {
                'лосось': [
                    'хорошая рыба'
                ],
                'килька': [
                    'консервы'
                ]
            },
            'default': 'дефолтный тунец'
        }
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.tokenized_elements_list_pymorphy = [
            {'lemma': 'мой'},
            {'lemma': 'дядя'},
            {'lemma': 'самый'},
            {'lemma': 'честный'},
            {'lemma': 'правило'},
        ]

        filler = IntersectionFieldFiller(items)
        result = filler.extract(text_preprocessing_result, None)

        self.assertEqual(expected, result)


class TestIntersectionOriginalTextFiller(TestCase):
    @patch('smart_kit.configs.get_app_config')
    def test_1(self, mock_get_app_config):
        patch_get_app_config(mock_get_app_config)
        items = {
            'cases': {
                'лосось': [
                    'хорошая рыба'
                ],
                'килька': [
                    'консервы'
                ]
            }
        }
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.original_text = 'всего хорошего и спасибо за рыбу'

        filler = IntersectionOriginalTextFiller(items)
        result = filler.extract(text_preprocessing_result, None)

        self.assertIsNone(result)

    @patch('smart_kit.configs.get_app_config')
    def test_2(self, mock_get_app_config):
        expected = 'лосось'
        patch_get_app_config(mock_get_app_config)
        items = {
            'cases': {
                'лосось': [
                    'хорошая рыба'
                ],
                'килька': [
                    'консервы'
                ]
            }
        }
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.original_text = 'всего хорошая и спасибо за рыба'

        filler = IntersectionOriginalTextFiller(items)
        result = filler.extract(text_preprocessing_result, None)

        self.assertEqual(expected, result)

    @patch('smart_kit.configs.get_app_config')
    def test_3(self, mock_get_app_config):
        patch_get_app_config(mock_get_app_config)
        items = {
            'cases': {
                'лосось': [
                    'хорошая рыба'
                ],
                'килька': [
                    'консервы'
                ]
            },
            'exceptions': {
                'лосось': 'не хорошая рыба'
            },
        }
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.original_text = 'не это хорошая рыба'

        filler = IntersectionOriginalTextFiller(items)
        result = filler.extract(text_preprocessing_result, None)

        self.assertIsNone(result)



import cmd
import json
import os
import pprint
from typing import Dict, Optional, Tuple, Any
from functools import cached_property

from core.descriptions.descriptions import Descriptions
from core.message.from_message import SmartAppFromMessage
from smart_kit.compatibility.commands import combine_commands
from smart_kit.testing import suite
from smart_kit.testing import utils


class CLInterface(cmd.Cmd):
    intro = "Привет!\t Введите help или ? для вызова списка команд.\n"
    prompt = "> "
    VPS_KEYS = ["answer", "pronounceText", "items", "suggestions"]

    def __init__(
            self, configs_path, secret_path, settings_cls, references_path,
            resources_cls, model_cls, test_case_cls, dialogue_manager_cls, user_cls, parametrizer_cls,
            from_msg_cls, app_name,
    ):
        super().__init__()
        self.configs_path = configs_path
        self.references_path = references_path
        self.settings = settings_cls(config_path=self.configs_path, secret_path=secret_path,
                                     references_path=self.references_path)

        self.environment = utils.Environment()
        self.lt_settings = utils.Settings()

        self.user_data = None

        self.__resources_cls = resources_cls
        self.__model_cls = model_cls
        self.__dialogue_manager_cls = dialogue_manager_cls
        self.__user_cls = user_cls
        self.__parametrizer_cls = parametrizer_cls
        self.__from_msg_cls = from_msg_cls
        self.__test_case_cls = test_case_cls

        predefined_fields_storage_path = os.path.join(self.references_path, "./predefined_fields_storage.json")
        if os.path.exists(predefined_fields_storage_path):
            with open(str(predefined_fields_storage_path), "r") as f:
                self.storaged_predefined_fields = json.load(f)
        else:
            self.storaged_predefined_fields = {}

    def after_process_message(self, message) -> Optional[str]:
        callback = getattr(self, f"on_{message.name.lower()}", lambda *args, **kwargs: None)
        return callback(message)

    @cached_property
    def app_model(self):
        return self.__model_cls(self.resources, self.__dialogue_manager_cls, custom_settings=self.settings)

    @cached_property
    def resources(self):
        source = self.settings.get_source()

        return self.__resources_cls(source, self.references_path, self.settings)

    @cached_property
    def available_scenarios(self):
        return list(Descriptions(self.resources.registered_repositories)["scenarios"].keys())

    @staticmethod
    def format_answer_value(ans):
        resp = ""
        if ans["messageName"] == "ANSWER_TO_USER":
            params = ans["payload"]
            for k in CLInterface.VPS_KEYS:
                if params.get(k):
                    resp += "{}: {}\n".format(k, params[k])
            if not resp:
                resp += "answer: <пустой ответ>\n"
            for key in filter(lambda key: key not in CLInterface.VPS_KEYS, params.keys()):
                if isinstance(params[key], list):
                    resp += "\t" + " | ".join("[{:^40}]".format(item) for item in params[key]) + "\n"
                else:
                    resp += "Other node {}:\t{}\n".format(key, params[key])
        else:
            resp += json.dumps(ans)
        return resp.strip()

    def do_show_envs(self, _input: str):
        self._show_envs()

    def do_set(self, _input: str):
        try:
            var, value = _input.split(' ', maxsplit=1)
        except ValueError:
            print("Нужно указать переменную и значение. Недостаточно данных")
            self._show_envs()
            return

        if var in self.environment:
            setattr(self.environment, var, value)
        elif var in self.lt_settings:
            setattr(self.lt_settings, var, value)
        else:
            print(f"Неизвестное значение: {var}")
            return

        print(f"{var} = {value}")

    def do_run(self, file_name: str):
        try:
            suite.run_testfile(
                os.getcwd(),
                file_name,
                self.app_model,
                self.settings,
                self.__user_cls,
                self.__parametrizer_cls,
                self.__from_msg_cls,
                self.__test_case_cls,
                self.storaged_predefined_fields,
                interactive=True,
            )
        except FileNotFoundError:
            print("Нужно указать относительный путь до json файла.")

    def complete_set(self, text, line, begidx, endidx):
        names = list(self.environment) + list(self.lt_settings)
        return list(filter(lambda t: t.startswith(text), names))

    def do_show_scenarios(self, _input: str):
        for scen_name in self.available_scenarios:
            print(f"\t{scen_name}")

    def preloop(self):
        if len(self.available_scenarios) == 0:
            print("Нет доступных сценариев.")
            exit(0)

        self.environment.intent = self.available_scenarios[0]
        print("Текущий сценарий: ", self.environment.intent)

    def process_message(self, raw_message: Dict, headers: tuple = ()) -> Tuple[Any, list]:
        from smart_kit.configs import get_app_config
        masking_fields = self.settings["template_settings"].get("masking_fields")
        message = SmartAppFromMessage(raw_message, headers=headers, masking_fields=masking_fields,
                                      validators=get_app_config().FROM_MSG_VALIDATORS)
        user = self.__user_cls(self.environment.user_id, message, self.user_data, self.settings,
                               self.app_model.scenario_descriptions,
                               self.__parametrizer_cls, load_error=False)
        answers = self.app_model.answer(message, user)
        return user, answers or []

    def default(self, _input: str):
        self.environment.text = _input
        self.environment.message_id += 1

        if self.lt_settings.DISPLAY_RAW:
            pprint.pprint(self.environment.as_dict)

        user, answers = self.process_message(self.environment.as_dict)
        self.user_data = user.raw_str

        answers = combine_commands(answers, user)
        for answer in answers:
            respond = self.after_process_message(answer)
            if respond:
                _, new_answers = self.process_message(
                    json.loads(respond),
                    (('app_callback_id', answer.request_data['app_callback_id'].encode()),),
                )
                answers.extend(new_answers or [])

        if not answers:
            print("Нет ответа")
            return

        for answer in answers:
            ans = answer.raw
            print("{}\nrequest type: {}\n".format(self.format_answer_value(ans), answer.request_type))
            if ans["messageName"] == "ANSWER_TO_USER":
                params = ans["payload"]
                intent = params.get("intent")
                if intent:
                    self.environment.intent = intent
            if self.lt_settings.DISPLAY_RAW:
                pprint.pprint(answer.raw)

    def run_local_testing(self):
        return self.cmdloop()

    def _show_envs(self):
        print("Доступные переменные:")
        for var_name, var_value in self.environment.items():
            print(f"\t{var_name} {var_value}")
        print("Доступные настройки:")
        for var_name, var_value in self.lt_settings.items():
            print(f"\t{var_name} {var_value}")
        # DRY!



# coding: utf-8


class LastScenarios:
    def __init__(self, items, description, user):
        self.scenarios_names = items or []
        self.description = description
        self.user = user

    def add(self, scenario_name, text_preprocessing_result):
        if not self.scenarios_names or scenario_name != self.scenarios_names[-1]:
            self._clear(scenario_name, text_preprocessing_result)
            self.scenarios_names.append(scenario_name)

    def _clear(self, scenario_name, text_preprocessing_result):
        count = self.description.get_count(text_preprocessing_result, self.user)
        if len(self.scenarios_names) >= count:
            last_scenario = self.scenarios_names[0]
            self.delete(last_scenario)
            self.user.forms.clear_form(last_scenario)
        if scenario_name in self.scenarios_names:
            self.scenarios_names.remove(scenario_name)

    def delete(self, scenario_name):
        if scenario_name in self.scenarios_names:
            self.scenarios_names.remove(scenario_name)


    @property
    def last_scenario_name(self):
        return self.scenarios_names[-1] if self.scenarios_names else None

    def clear_all(self):
        self.scenarios_names = []

    def expire(self):
        to_remove = []
        scenario_descriptions = self.user.descriptions["scenarios"]
        forms = self.user.forms
        for scenario_id in self.scenarios_names:
            if scenario_id not in scenario_descriptions:
                to_remove.append(scenario_id)
                continue

            scenario_description = scenario_descriptions[scenario_id]

            if not hasattr(scenario_description, "form_type"):
                to_remove.append(scenario_id)
                continue

            form_key = scenario_description.form_type
            if not forms[form_key]:
                to_remove.append(scenario_id)

        for scenario_id in to_remove:
            self.delete(scenario_id)

    @property
    def raw(self):
        return self.scenarios_names



import json
import typing
import os

import asyncio
import concurrent.futures
import aiohttp
import aiohttp.web

import scenarios.logging.logger_constants as log_const
from core.db_adapter.db_adapter import DBAdapterException, db_adapter_factory
from core.logging.logger_utils import log
from core.message.from_message import SmartAppFromMessage
from core.utils.stats_timer import StatsTimer
from smart_kit.message.smartapp_to_message import SmartAppToMessage
from smart_kit.start_points.main_loop_http import BaseHttpMainLoop
from core.monitoring.monitoring import monitoring


class AIOHttpMainLoop(BaseHttpMainLoop):
    def __init__(self, *args, **kwargs):
        self.app = aiohttp.web.Application()
        self.app.add_routes([aiohttp.web.route('*', '/{tail:.*}', self.iterate)])
        super().__init__(*args, **kwargs)
        max_workers = self.settings["template_settings"].get("max_workers", (os.cpu_count() or 1) * 5)
        self.pool = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)

    async def async_init(self):
        await self.db_adapter.connect()

    def get_db(self):
        db_adapter = db_adapter_factory(self.settings["template_settings"].get("db_adapter", {}))
        self.app.on_cleanup.append(self.close_db)
        return db_adapter

    # noinspection PyMethodMayBeStatic
    async def close_db(self, app):
        app["database"].cancel()
        await app["database"]

    async def load_user(self, db_uid, message):
        db_data = None
        load_error = False
        try:
            if self.db_adapter.IS_ASYNC:
                db_data = await self.db_adapter.get(db_uid)
            else:
                db_data = self.db_adapter.get(db_uid)
        except (DBAdapterException, ValueError):
            log("Failed to get user data", params={log_const.KEY_NAME: log_const.FAILED_DB_INTERACTION,
                                                   log_const.REQUEST_VALUE: message.as_str}, level="ERROR")
            load_error = True
            monitoring.counter_load_error(self.app_name)
        return self.user_cls(
            message.uid,
            message=message,
            db_data=db_data,
            settings=self.settings,
            descriptions=self.model.scenario_descriptions,
            parametrizer_cls=self.parametrizer_cls,
            load_error=load_error
        )

    async def save_user(self, db_uid, user, message):
        no_collisions = True
        if user.do_not_save:
            log("User %(uid)s will not saved", user=user, params={"uid": user.id,
                                                                  log_const.KEY_NAME: "user_will_not_saved"})
        else:

            no_collisions = True
            try:
                str_data = user.raw_str

                if self.db_adapter.IS_ASYNC:
                    if user.initial_db_data and self.user_save_check_for_collisions:
                        no_collisions = await self.db_adapter.replace_if_equals(
                            db_uid,
                            sample=user.initial_db_data,
                            data=str_data
                        )
                    else:
                        await self.db_adapter.save(db_uid, str_data)
                else:
                    if user.initial_db_data and self.user_save_check_for_collisions:
                        no_collisions = self.db_adapter.replace_if_equals(
                            db_uid,
                            sample=user.initial_db_data,
                            data=str_data
                        )
                    else:
                        self.db_adapter.save(db_uid, str_data)
            except (DBAdapterException, ValueError):
                log("Failed to set user data", params={log_const.KEY_NAME: log_const.FAILED_DB_INTERACTION,
                                                       log_const.REQUEST_VALUE: message.as_str}, level="ERROR")
                monitoring.counter_save_error(self.app_name)
            if not no_collisions:
                monitoring.counter_save_collision(self.app_name)
        return no_collisions

    def run(self):
        aiohttp_config = self.settings["aiohttp"]
        if not aiohttp_config:
            log("aiohttp.yml is empty or missing. Server will be started with default parameters", level="WARN")
        asyncio.get_event_loop().run_until_complete(self.async_init())
        aiohttp.web.run_app(app=self.app, **aiohttp_config)

    def stop(self, signum, frame):
        pass

    async def handle_message(self, message: SmartAppFromMessage) -> typing.Tuple[int, str, SmartAppToMessage]:
        if not message.validate():
            return 400, "BAD REQUEST", SmartAppToMessage(self.BAD_REQUEST_COMMAND, message=message, request=None)

        answer, stats = await self.process_message(message)
        if not answer:
            return 204, "NO CONTENT", SmartAppToMessage(self.NO_ANSWER_COMMAND, message=message, request=None)

        answer_message = SmartAppToMessage(
            answer, message, request=None,
            validators=self.to_msg_validators)
        if answer_message.validate():
            return 200, "OK", answer_message
        else:
            return 500, "BAD ANSWER", SmartAppToMessage(self.BAD_ANSWER_COMMAND, message=message, request=None)

    async def process_message(self, message: SmartAppFromMessage, *args, **kwargs):
        stats = ""
        log("INCOMING DATA: %(masked_message)s",
            params={log_const.KEY_NAME: "incoming_policy_message", "masked_message": message.masked_value})
        db_uid = message.db_uid

        with StatsTimer() as load_timer:
            user = await self.load_user(db_uid, message)
        stats += "Loading time: {} msecs\n".format(load_timer.msecs)
        with StatsTimer() as script_timer:
            commands = await self.app.loop.run_in_executor(self.pool, self.model.answer, message, user)
            if commands:
                answer = self._generate_answers(user, commands, message)
            else:
                answer = None

        stats += "Script time: {} msecs\n".format(script_timer.msecs)
        with StatsTimer() as save_timer:
            await self.save_user(db_uid, user, message)
        stats += "Saving time: {} msecs\n".format(save_timer.msecs)
        log(stats, params={log_const.KEY_NAME: "timings"})
        self.postprocessor.postprocess(user, message)
        return answer, stats

    async def iterate(self, request: aiohttp.web.Request):
        headers = self._get_headers(request.headers)
        body = await request.text()
        message = SmartAppFromMessage(json.loads(body), headers=headers, headers_required=False,
                                      validators=self.from_msg_validators)

        status, reason, answer = await self.handle_message(message)

        return aiohttp.web.json_response(
            status=status, reason=reason, data=answer.as_dict,
            headers=self._get_outgoing_headers(headers, answer.command)
        )



# coding: utf-8
from typing import Dict, Callable, Any

from core.descriptions.descriptions_items import DescriptionsItems
from core.model.registered import Registered
from core.repositories.base_repository import BaseRepository

registered_description_factories = Registered()


def default_description_factory(x):
    return x


class Descriptions:
    def __init__(self, registered_repositories: Dict[str, BaseRepository]) -> None:
        self.registered_repositories: Dict[str, BaseRepository] = registered_repositories
        self._descriptions: dict = {}

    def __getitem__(self, key: str) -> Any:
        description_item = self._descriptions.get(key)
        if description_item is None:
            repository: BaseRepository = self.registered_repositories[key]
            factory: Callable = registered_description_factories.get(key, default_description_factory)
            description_item = factory(repository.data)
            self._descriptions[key] = description_item
        return description_item

    def __setitem__(self, key: str, description_item: DescriptionsItems) -> None:
        self._descriptions[key] = description_item



import os
import shutil

import smart_kit
from smart_kit.management.base import Manager, AppCommand


class SmartKitManager(Manager):
    def default(self, argv):
        return self.execute_command("help", *argv[2:])


class VersionCommand(AppCommand):
    def execute(self, *args, **kwargs):
        print(smart_kit.__version__)


class CreateAppCommand(AppCommand):
    """
        Create app folder at given path.
    """

    TEMPLATE_FOLDER = "template"

    def __init__(self, extensions: tuple = None):
        if extensions:
            self.extensions = extensions
        else:
            self.extensions = (".py", ".yml")

    def execute(self, *args, **kwargs):
        app_name = args[0]
        new_app_path = os.path.join(os.getcwd(), app_name)
        try:
            os.makedirs(new_app_path)
        except FileExistsError:
            raise
        template_dir = os.path.join(smart_kit.__path__[0], self.TEMPLATE_FOLDER)
        prefix_length = len(template_dir) + 1
        for root, dirs, files in os.walk(template_dir):

            relative_dir = root[prefix_length:]

            if relative_dir:
                target_dir = os.path.join(new_app_path, relative_dir)
                os.makedirs(target_dir, exist_ok=True)

            for filename in files:
                old_path = os.path.join(root, filename)
                new_path = os.path.join(
                    new_app_path, relative_dir, filename.replace("-tpl", "") if "-tpl" in filename else filename)
                if new_path.endswith(self.extensions):
                    with open(old_path, encoding='utf-8') as template_file:
                        content = template_file.read()
                    content = content.replace("{{app_name}}", app_name)
                    with open(new_path, 'w', encoding='utf-8') as new_file:
                        new_file.write(content)
                else:
                    shutil.copyfile(old_path, new_path)

        print(
            f"{new_app_path} - app created"
        )



# coding: utf-8


class LazyItems:
    def __init__(self, items, descriptions, user, factory):
        self._descriptions = descriptions
        self._factory = factory
        self._raw_items = items or {}
        self._items = dict()
        self._user = user
        self._clear_removed_items()

    def _clear_removed_items(self):
        keys_for_delete = []
        for key in self._raw_items:
            if not key in self._descriptions:
                keys_for_delete.append(key)
        for key in keys_for_delete:
            self._raw_items.pop(key)

    def __getitem__(self, description):
        if hasattr(description, "id"):
            id = description.id
        else:
            id = description
            description = self._descriptions[id]
        existed_item = self._items.get(description)
        if existed_item is None:
            raw_data = self._raw_items.get(id)
            existed_item = self._build_factory(description, raw_data)
            self._items[description] = existed_item
        return existed_item

    def _build_factory(self, description, raw_data):
        return self._factory(raw_data, description, self._user)

    def __setitem__(self, description, value):
        self._items[description] = value

    def __iter__(self):
        return iter(self._descriptions[key] for key in self._descriptions)

    def remove_item(self, descr_id):
        description = self._descriptions[descr_id]
        if description in self._items:
            self._items.pop(description)
        if descr_id in self._raw_items:
            self._raw_items.pop(descr_id)

    @property
    def descriptions(self):
        return self._descriptions

    @property
    def raw(self):
        for i in self._items:
            item = self._items[i]
            value = getattr(item, "raw", item)
            if value is not None:
                self._raw_items[i.id] = value
            else:
                if i.id in self._raw_items:
                    self._raw_items.pop(i.id)
        return self._raw_items



# coding: utf-8
import json
import unittest
import uuid
from unittest.mock import Mock, MagicMock, patch

from core.basic_models.actions.basic_actions import Action, DoingNothingAction, action_factory, RequirementAction, \
    actions, ChoiceAction, ElseAction, CompositeAction, NonRepeatingAction
from core.basic_models.actions.client_profile_actions import GiveMeMemoryAction, RememberThisAction
from core.basic_models.actions.command import Command
from core.basic_models.actions.counter_actions import CounterIncrementAction, CounterDecrementAction, \
    CounterClearAction, CounterSetAction, CounterCopyAction
from core.basic_models.actions.external_actions import ExternalAction
from core.basic_models.actions.push_action import PushAction
from core.basic_models.actions.string_actions import StringAction, AfinaAnswerAction, SDKAnswer, \
    SDKAnswerToUser, NodeAction
from core.basic_models.answer_items.answer_items import SdkAnswerItem, items_factory, answer_items, BubbleText, \
    ItemCard, PronounceText, SuggestText, SuggestDeepLink
from core.basic_models.requirement.basic_requirements import requirement_factory, Requirement, requirements
from core.model.registered import registered_factories
from core.unified_template.unified_template import UnifiedTemplate, UNIFIED_TEMPLATE_TYPE_NAME
from smart_kit.request.kafka_request import SmartKitKafkaRequest
from smart_kit.utils.picklable_mock import PicklableMock, PicklableMagicMock


class MockParametrizer:
    def __init__(self, user, items=None):
        self.items = items or {}
        self.user = user
        self.data = items.get("data") or {}
        self.filter = items.get("filter") or False

    def collect(self, text_preprocessing_result=None, filter_params=None):
        data = {
            "payload": self.user.message.payload
        }
        data.update(self.data)
        if self.filter:
            data.update({"filter": "filter_out"})
        return data


class MockAction:
    def __init__(self, items=None):
        items = items or {}
        self.result = items.get("result")

    def run(self, user, text_preprocessing_result, params=None):
        return self.result or ["test action run"]


class UserMockAction:
    def __init__(self, items=None):
        items = items or {}
        self.result = items.get("result")
        self.done = False

    def run(self, user, text_preprocessing_result, params=None):
        self.done = True


class MockRequirement:
    def __init__(self, items):
        self.result = items.get("result")

    def check(self, text_preprocessing_result, user, params):
        return self.result


class MockSimpleParametrizer:
    def __init__(self, user, items=None):
        self.user = user
        self.data = items.get("data")

    def collect(self, text_preprocessing_result, filter_params=None):
        return self.data


class ActionTest(unittest.TestCase):
    def test_nodes_1(self):
        items = {"nodes": {"answer": "test"}}
        action = NodeAction(items)
        nodes = action.nodes
        action_key = list(nodes.keys())[0]
        self.assertEqual(action_key, 'answer')
        self.assertIsInstance(nodes[action_key], UnifiedTemplate)

    def test_nodes_2(self):
        items = {}
        action = NodeAction(items)
        nodes = action.nodes
        self.assertEqual(nodes, {})

    def test_base(self):
        items = {"nodes": "test"}
        action = Action(items)
        result = action.run(None, None)
        self.assertEqual(result, [])

    def test_external(self):
        items = {"action": "test_action_key"}
        action = ExternalAction(items)
        user = PicklableMock()
        user.descriptions = {"external_actions": {"test_action_key": MockAction()}}
        self.assertEqual(action.run(user, None), ["test action run"])

    def test_doing_nothing_action(self):
        items = {"nodes": {"answer": "test"}, "command": "test_name"}
        action = DoingNothingAction(items)
        result = action.run(None, None)
        self.assertIsInstance(result, list)
        command = result[0]
        self.assertIsInstance(command, Command)
        self.assertEqual(command.name, "test_name")
        self.assertEqual(command.payload, {"answer": "test"})

    def test_requirement_action(self):
        registered_factories[Requirement] = requirement_factory
        requirements["test"] = MockRequirement
        registered_factories[Action] = action_factory
        actions["test"] = MockAction
        items = {"requirement": {"type": "test", "result": True}, "action": {"type": "test"}}
        action = RequirementAction(items)
        self.assertIsInstance(action.requirement, MockRequirement)
        self.assertIsInstance(action.internal_item, MockAction)
        self.assertEqual(action.run(None, None), ["test action run"])
        items = {"requirement": {"type": "test", "result": False}, "action": {"type": "test"}}
        action = RequirementAction(items)
        result = action.run(None, None)
        self.assertEqual([], result)

    def test_requirement_choice(self):
        items = {"requirement_actions": [
            {"requirement": {"type": "test", "result": False}, "action": {"type": "test", "result": ["action1"]}},
            {"requirement": {"type": "test", "result": True}, "action": {"type": "test", "result": ["action2"]}}
        ]}
        choice_action = ChoiceAction(items)
        self.assertIsInstance(choice_action.items, list)
        self.assertIsInstance(choice_action.items[0], RequirementAction)
        result = choice_action.run(None, None)
        self.assertEqual(result, ["action2"])

    def test_requirement_choice_else(self):
        items = {
            "requirement_actions": [
                {"requirement": {"type": "test", "result": False}, "action": {"type": "test", "result": "action1"}},
                {"requirement": {"type": "test", "result": False}, "action": {"type": "test", "result": "action2"}},
            ],
            "else_action": {"type": "test", "result": ["action3"]}
        }
        choice_action = ChoiceAction(items)
        self.assertIsInstance(choice_action.items, list)
        self.assertIsInstance(choice_action.items[0], RequirementAction)
        result = choice_action.run(None, None)
        self.assertEqual(result, ["action3"])

    def test_string_action(self):
        expected = [Command("cmd_id", {"item": "template", "params": "params"})]
        user = PicklableMagicMock()
        template = PicklableMock()
        template.get_template = Mock(return_value=["nlpp.payload.personInfo.identityCard"])
        user.descriptions = {"render_templates": template}
        params = {"params": "params"}
        user.parametrizer = MockSimpleParametrizer(user, {"data": params})
        items = {"command": "cmd_id",
                 "nodes":
                     {"item": "template", "params": "{{params}}"}}
        action = StringAction(items)
        result = action.run(user, None)
        self.assertEqual(expected[0].name, result[0].name)
        self.assertEqual(expected[0].payload, result[0].payload)

    def test_else_action_if(self):
        registered_factories[Requirement] = requirement_factory
        requirements["test"] = MockRequirement
        registered_factories[Action] = action_factory
        actions["test"] = MockAction
        user = PicklableMock()
        items = {
            "requirement": {"type": "test", "result": True},
            "action": {"type": "test", "result": ["main_action"]},
            "else_action": {"type": "test", "result": ["else_action"]}
        }
        action = ElseAction(items)
        self.assertEqual(action.run(user, None), ["main_action"])

    def test_else_action_else(self):
        registered_factories[Requirement] = requirement_factory
        requirements["test"] = MockRequirement
        registered_factories[Action] = action_factory
        actions["test"] = MockAction
        user = PicklableMock()
        items = {
            "requirement": {"type": "test", "result": False},
            "action": {"type": "test", "result": ["main_action"]},
            "else_action": {"type": "test", "result": ["else_action"]}
        }
        action = ElseAction(items)
        self.assertEqual(action.run(user, None), ["else_action"])

    def test_else_action_no_else_if(self):
        registered_factories[Requirement] = requirement_factory
        requirements["test"] = MockRequirement
        registered_factories[Action] = action_factory
        actions["test"] = MockAction
        user = PicklableMock()
        items = {
            "requirement": {"type": "test", "result": True},
            "action": {"type": "test", "result": ["main_action"]},
        }
        action = ElseAction(items)
        self.assertEqual(action.run(user, None), ["main_action"])

    def test_else_action_no_else_else(self):
        registered_factories[Requirement] = requirement_factory
        requirements["test"] = MockRequirement
        registered_factories[Action] = action_factory
        actions["test"] = MockAction
        user = PicklableMock()
        items = {
            "requirement": {"type": "test", "result": False},
            "action": {"type": "test", "result": ["main_action"]},
        }
        action = ElseAction(items)
        result = action.run(user, None)
        self.assertEqual([], result)

    def test_composite_action(self):
        registered_factories[Action] = action_factory
        actions["action_mock"] = MockAction
        user = PicklableMock()
        items = {
            "actions": [
                {"type": "action_mock"},
                {"type": "action_mock"}
            ]
        }
        action = CompositeAction(items)
        result = action.run(user, None)
        self.assertEqual(['test action run', 'test action run'], result)

    def test_node_action_support_templates(self):
        params = {
            "markup": "italic",
            "email": "heyho@sberbank.ru",
            "name": "Buratino"
        }
        items = {
            "support_templates": {
                "markup": "{%if markup=='italic'%}i{% else %}b{% endif %}",
                "email": "{%if email%}<{{markup}}>Email: {{email}}</{{markup}}>\n{% endif %}",
                "name": "{%if name%}<{{markup}}>Name: {{name}}</{{markup}}>\n{% endif %}",
                "result": "{{email}}{{name}}"
            },
            "nodes": {
                "answer": "{{result|trim}}"
            }
        }
        expected = "<i>Email: heyho@sberbank.ru</i>\n<i>Name: Buratino</i>"

        action = StringAction(items)
        for template_key, template in action.support_templates.items():
            self.assertIsInstance(template, UnifiedTemplate)
        user = PicklableMagicMock()
        user.parametrizer = MockSimpleParametrizer(user, {"data": params})
        output = action.run(user=user, text_preprocessing_result=None)[0].payload["answer"]
        self.assertEqual(output, expected)

    def test_string_action_support_templates(self):
        params = {
            "answer_text": "some_text",
            "buttons_number": 3
        }
        items = {
            "nodes": {
                "answer": "{{ answer_text }}",
                "buttons": {
                    "type": UNIFIED_TEMPLATE_TYPE_NAME,
                    "template": "{{range(buttons_number)|list}}",
                    "loader": "json"
                }
            }
        }
        expected = {
            "answer": "some_text",
            "buttons": [0, 1, 2]
        }
        action = StringAction(items)
        user = PicklableMagicMock()
        user.parametrizer = MockSimpleParametrizer(user, {"data": params})
        output = action.run(user=user, text_preprocessing_result=None)[0].payload
        self.assertEqual(output, expected)

    def test_push_action(self):
        params = {
            "day_time": "morning",
            "deep_link_url": "some_url",
            "icon_url": "some_icon_url"
        }
        settings = {"template_settings": {"project_id": "project_id"}}
        items = {
            "request_data": {
                "kafka_extraHeaders": {
                    "request-id": "{{ 1 }}"
                }
            },
            "content": {
                "notificationHeader": "{% if day_time == 'morning' %}Время завтракать!{% else %}Хотите что нибудь заказать?{% endif %}",
                "fullText": "В нашем магазине большой ассортимент{% if day_time == 'evening' %}. Успей заказать!{% endif %}",
                "mobileAppParameters": {
                    "DeeplinkAndroid": "{{ deep_link_url }}",
                    "DeeplinkIos": "{{ deep_link_url }}",
                    "Logo": "{{ icon_url }}"
                }
            }
        }
        user = PicklableMagicMock()

        expected = {
            "surface": "COMPANION",
            "projectId": "project_id",
            "clientId": user.message.sub,
            "content": {
                "notificationHeader": "Время завтракать!",
                "fullText": "В нашем магазине большой ассортимент",
                "mobileAppParameters": {
                    "DeeplinkAndroid": "some_url",
                    "DeeplinkIos": "some_url",
                    "Logo": "some_icon_url"
                }
            }
        }
        action = PushAction(items)
        user.parametrizer = MockSimpleParametrizer(user, {"data": params})
        user.settings = settings
        command = action.run(user=user, text_preprocessing_result=None)[0]
        self.assertEqual(command.raw, expected)
        # проверяем наличие кастомных хэдеров для сервиса пушей
        self.assertTrue(SmartKitKafkaRequest.KAFKA_EXTRA_HEADERS in command.request_data)
        headers = command.request_data.get(SmartKitKafkaRequest.KAFKA_EXTRA_HEADERS)
        self.assertTrue('request-id' in headers)
        self.assertTrue('sender-id' in headers)
        self.assertEqual(headers.get('request-id'), "1")
        self.assertTrue(uuid.UUID(headers.get('sender-id'), version=3))


class NonRepeatingActionTest(unittest.TestCase):
    def setUp(self):
        self.expected = [PicklableMock()]
        self.expected1 = [PicklableMock()]
        self.action = NonRepeatingAction({"actions": [{"type": "action_mock",
                                                       "result": self.expected},
                                                      {"type": "action_mock",
                                                       "result": self.expected1}
                                                      ],
                                          "last_action_ids_storage": "last_action_ids_storage"})
        self.user = PicklableMagicMock()
        registered_factories[Action] = action_factory
        actions["action_mock"] = MockAction

    def test_run_available_indexes(self):
        self.user.last_action_ids["last_action_ids_storage"].get_list.side_effect = [[0]]
        result = self.action.run(self.user, None)
        self.user.last_action_ids["last_action_ids_storage"].add.assert_called_once()
        self.assertEqual(result, self.expected1)

    def test_run_no_available_indexes(self):
        self.user.last_action_ids["last_action_ids_storage"].get_list.side_effect = [[0, 1]]
        result = self.action.run(self.user, None)
        self.assertEqual(result, self.expected)


class CounterIncrementActionTest(unittest.TestCase):
    def test_run(self):
        user = PicklableMock()
        counter = PicklableMock()
        counter.inc = PicklableMock()
        user.counters = {"test": counter}
        items = {"key": "test"}
        action = CounterIncrementAction(items)
        action.run(user, None)
        user.counters["test"].inc.assert_called_once()


class CounterDecrementActionTest(unittest.TestCase):
    def test_run(self):
        user = PicklableMock()
        counter = PicklableMock()
        counter.dec = PicklableMock()
        user.counters = {"test": counter}
        items = {"key": "test"}
        action = CounterDecrementAction(items)
        action.run(user, None)
        user.counters["test"].dec.assert_called_once()


class CounterClearActionTest(unittest.TestCase):
    def test_run(self):
        user = PicklableMock()
        user.counters = PicklableMock()
        user.counters.inc = PicklableMock()
        items = {"key": "test"}
        action = CounterClearAction(items)
        action.run(user, None)
        user.counters.clear.assert_called_once()


class CounterSetActionTest(unittest.TestCase):
    def test_run(self):
        user = PicklableMock()
        counter = PicklableMock()
        counter.inc = PicklableMock()
        counters = {"test": counter}
        user.counters = counters
        items = {"key": "test"}
        action = CounterSetAction(items)
        action.run(user, None)
        user.counters["test"].set.assert_called_once()


class CounterCopyActionTest(unittest.TestCase):
    def test_run(self):
        user = PicklableMock()
        counter_src = PicklableMock()
        counter_src.value = 10
        counter_dst = PicklableMock()
        user.counters = {"src": counter_src, "dst": counter_dst}
        items = {"source": "src", "destination": "dst"}
        action = CounterCopyAction(items)
        action.run(user, None)
        user.counters["dst"].set.assert_called_once_with(user.counters["src"].value,
                                                         action.reset_time, action.time_shift)


class AfinaAnswerActionTest(unittest.TestCase):
    def test_typical_answer(self):
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        expected = [MagicMock(_name="ANSWER_TO_USER", raw={'messageName': 'ANSWER_TO_USER',
                                                           'payload': {'answer': 'a1'}})]
        items = {
            "nodes": {
                "answer": ["a1", "a1", "a1"],
            }
        }
        action = AfinaAnswerAction(items)

        result = action.run(user, None)
        self.assertEqual(expected[0]._name, result[0].name)
        self.assertEqual(expected[0].raw, result[0].raw)

    def test_typical_answer_with_other(self):
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        expected = [MagicMock(_name="ANSWER_TO_USER", raw={'messageName': 'ANSWER_TO_USER',
                                                           'payload': {'answer': 'a1',
                                                                       "pronounce_text": 'pt2',
                                                                       "picture": "1.jpg"}})]
        items = {
            "nodes": {
                "answer": ["a1", "a1", "a1"],
                "pronounce_text": ["pt2"],
                "picture": ["1.jpg", "1.jpg", "1.jpg"]
            }
        }
        action = AfinaAnswerAction(items)

        result = action.run(user, None)
        self.assertEqual(expected[0]._name, result[0].name)
        self.assertEqual(expected[0].raw, result[0].raw)

    def test_typical_answer_with_pers_info(self):
        expected = [MagicMock(_name="ANSWER_TO_USER", raw={'messageName': 'ANSWER_TO_USER',
                                                           'payload': {'answer': 'Ivan Ivanov'}})]
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        user.message = PicklableMock()
        user.message.payload = {"personInfo": {"name": "Ivan Ivanov"}}
        items = {"nodes": {"answer": ["{{payload.personInfo.name}}"]}}
        action = AfinaAnswerAction(items)
        result = action.run(user, None)
        self.assertEqual(expected[0]._name, result[0].name)
        self.assertEqual(expected[0].raw, result[0].raw)

    def test_items_empty(self):
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        template = PicklableMock()
        template.get_template = Mock(return_value=[])
        user.descriptions = {"render_templates": template}
        items = None
        action = AfinaAnswerAction(items)
        result = action.run(user, None)
        self.assertEqual(result, [])

    def test__items_empty_dict(self):
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        template = PicklableMock()
        template.get_template = Mock(return_value=[])
        user.descriptions = {"render_templates": template}
        items = {}
        action = AfinaAnswerAction(items)
        result = action.run(user, None)
        self.assertEqual(result, [])


class CardAnswerActionTest(unittest.TestCase):
    def test_typical_answer(self):
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        user.message = PicklableMock()
        user.message.payload = {"personInfo": {"name": "Ivan Ivanov"}}
        items = {
            "type": "sdk_answer",
            "nodes": {
                "pronounceText": ["pronounceText1", "{{payload.personInfo.name}}"],
                "items": [
                  {
                    "bubble": {
                      "text": ["Text1", "Text2"]
                    }
                  },
                  {
                    "card": {
                      "type": "simple_list",
                      "header": "1 доллар США ",
                      "items": [
                        {
                          "title": "Купить",
                          "body": "67.73 RUR"
                        },
                        {
                          "title": "Продать",
                          "body": "64.56 RUR"
                        }
                      ],
                      "footer": "{{payload.personInfo.name}} Сбербанк Онлайн на сегодня 17:53 при обмене до 1000 USD"
                    }
                  }
                ],
                "suggestions": {
                     "buttons": [{
                        "title": ["Отделения"],
                        "action": {
                          "text": "Где ближайщие отделения сбера?",
                          "type": "text"
                        }
                     }]
                }
            }
        }
        exp1 = json.dumps({'messageName': 'ANSWER_TO_USER', 'payload': {'pronounceText': 'Ivan Ivanov', 'items': [{'bubble': {'text': 'Text2'}}, {'card': {'type': 'simple_list', 'header': '1 доллар США ', 'items': [{'title': 'Купить', 'body': '67.73 RUR'}, {'title': 'Продать', 'body': '64.56 RUR'}], 'footer': 'Ivan Ivanov Сбербанк Онлайн на сегодня 17:53 при обмене до 1000 USD'}}], 'suggestions': {'buttons': [{'title': 'Отделения', 'action': {'text': 'Где ближайщие отделения сбера?', 'type': 'text'}}]}}}, sort_keys=True)
        exp2 = json.dumps({'messageName': 'ANSWER_TO_USER', 'payload': {'pronounceText': 'pronounceText1', 'items': [{'bubble': {'text': 'Text1'}}, {'card': {'type': 'simple_list', 'header': '1 доллар США ', 'items': [{'title': 'Купить', 'body': '67.73 RUR'}, {'title': 'Продать', 'body': '64.56 RUR'}], 'footer': 'Ivan Ivanov Сбербанк Онлайн на сегодня 17:53 при обмене до 1000 USD'}}], 'suggestions': {'buttons': [{'title': 'Отделения', 'action': {'text': 'Где ближайщие отделения сбера?', 'type': 'text'}}]}}}, sort_keys=True)
        expect_arr = [exp1, exp2]

        nexp1 = json.dumps({'messageName': 'ANSWER_TO_USER', 'payload': {'pronounceText': 'Ivan Ivanov', 'items': [{'bubble': {'text': 'Text1'}}, {'card': {'type': 'simple_list', 'header': '1 доллар США ', 'items': [{'title': 'Купить', 'body': '67.73 RUR'}, {'title': 'Продать', 'body': '64.56 RUR'}], 'footer': 'Ivan Ivanov Сбербанк Онлайн на сегодня 17:53 при обмене до 1000 USD'}}], 'suggestions': {'buttons': [{'title': 'Отделения', 'action': {'text': 'Где ближайщие отделения сбера?', 'type': 'text'}}]}}}, sort_keys=True)
        nexp2 = json.dumps({'messageName': 'ANSWER_TO_USER', 'payload': {'pronounceText': 'pronounceText1', 'items': [{'bubble': {'text': 'Text2'}}, {'card': {'type': 'simple_list', 'header': '1 доллар США ', 'items': [{'title': 'Купить', 'body': '67.73 RUR'}, {'title': 'Продать', 'body': '64.56 RUR'}], 'footer': 'Ivan Ivanov Сбербанк Онлайн на сегодня 17:53 при обмене до 1000 USD'}}], 'suggestions': {'buttons': [{'title': 'Отделения', 'action': {'text': 'Где ближайщие отделения сбера?', 'type': 'text'}}]}}}, sort_keys=True)
        not_expect_arr = [nexp1, nexp2]

        for i in range(10):
            action = SDKAnswer(items)
            result = action.run(user, None)
            self.assertEqual("ANSWER_TO_USER", result[0].name)
            self.assertTrue(json.dumps(result[0].raw, sort_keys=True) in expect_arr)
            self.assertFalse(json.dumps(result[0].raw, sort_keys=True) in not_expect_arr)


    def test_typical_answer_without_items(self):
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        user.message = PicklableMock()
        user.message.payload = {"personInfo": {"name": "Ivan Ivanov"}}
        items = {
            "type": "sdk_answer",
            "nodes": {
                "pronounceText": ["pronounceText1", "pronounceText2"],
            }
        }
        exp1 = json.dumps({'messageName': 'ANSWER_TO_USER', 'payload': {'pronounceText': 'pronounceText1'}}, sort_keys=True)
        exp2 = json.dumps({'messageName': 'ANSWER_TO_USER', 'payload': {'pronounceText': 'pronounceText2'}}, sort_keys=True)
        exp_list = [exp1, exp2]
        for i in range(10):
            action = SDKAnswer(items)
            result = action.run(user, None)
            self.assertEqual("ANSWER_TO_USER", result[0].name)
            self.assertTrue(json.dumps((result[0].raw), sort_keys=True) in exp_list)

    def test_typical_answer_without_nodes(self):
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        user.message = PicklableMock()
        user.message.payload = {"personInfo": {"name": "Ivan Ivanov"}}
        items = {
                "type": "sdk_answer",
                "pronounceText": ["pronounceText1"],
                "suggestions": {
                    "buttons": [
                        {
                            "title": ["{{payload.personInfo.name}}", "отделения2"],
                            "action": {
                                "text": "отделения",
                                "type": "text"
                            }
                        },
                        {
                            "title": ["кредит1", "кредит2"],
                            "action": {
                                "text": "кредит",
                                "type": "text"
                            }
                        }
                    ]
                }
        }
        exp1 = json.dumps({'messageName': 'ANSWER_TO_USER', 'payload': {'pronounceText': 'pronounceText1', 'suggestions': {'buttons': [{'title': 'Ivan Ivanov', 'action': {'text': 'отделения', 'type': 'text'}}, {'title': 'кредит1', 'action': {'text': 'кредит', 'type': 'text'}}]}}}, sort_keys=True)
        exp2 = json.dumps({'messageName': 'ANSWER_TO_USER', 'payload': {'pronounceText': 'pronounceText1', 'suggestions': {'buttons': [{'title': 'отделения2', 'action': {'text': 'отделения', 'type': 'text'}}, {'title': 'кредит2', 'action': {'text': 'кредит', 'type': 'text'}}]}}}, sort_keys=True)
        expect_arr = [exp1, exp2]

        nexp1 = json.dumps({'messageName': 'ANSWER_TO_USER', 'payload': {'pronounceText': 'pronounceText1', 'suggestions': {'buttons': [{'title': 'Ivan Ivanov', 'action': {'text': 'отделения', 'type': 'text'}}, {'title': 'кредит2', 'action': {'text': 'кредит', 'type': 'text'}}]}}}, sort_keys=True)
        nexp2 = json.dumps({'messageName': 'ANSWER_TO_USER', 'payload': {'pronounceText': 'pronounceText1', 'suggestions': {'buttons': [{'title': 'отделения2', 'action': {'text': 'отделения', 'type': 'text'}}, {'title': 'кредит1', 'action': {'text': 'кредит', 'type': 'text'}}]}}}, sort_keys=True)
        not_expect_arr = [nexp1, nexp2]
        for i in range(10):
            action = SDKAnswer(items)
            result = action.run(user, None)
            self.assertEqual("ANSWER_TO_USER", result[0].name)
            self.assertTrue(json.dumps((result[0].raw), sort_keys=True) in expect_arr)
            self.assertFalse(json.dumps((result[0].raw), sort_keys=True) in not_expect_arr)


class SDKRandomAnswer(unittest.TestCase):
    def test_SDKItemAnswer_full(self):

        registered_factories[SdkAnswerItem] = items_factory
        answer_items["bubble_text"] = BubbleText
        answer_items["item_card"] = ItemCard
        answer_items["pronounce_text"] = PronounceText
        answer_items["suggest_text"] = SuggestText
        answer_items["suggest_deeplink"] = SuggestDeepLink

        registered_factories[Requirement] = requirement_factory
        requirements["test"] = MockRequirement
        requirements[None] = Requirement

        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        user.message = PicklableMock()
        user.message.payload = {"personInfo": {"name": "Ivan Ivanov"}}
        items = {
            "type": "sdk_answer_to_user",
            "static":
                {
                    "static_text": "st1",
                    "card1": {"cards_params": "a lot of params"},
                    "dl": "www.ww.w"
                },
            "random_choice": [
                {
                    "pron": "p1",
                    "txt": "{{payload.personInfo.name}}",
                    "title": "title1"
                },
                {
                    "pron": "p2",
                    "txt": "t2",
                    "title": "title2"
                }],
            "items":
                [
                    {
                        'type': "item_card",
                        "text": "txt",
                        "requirement": {"type": "test", "result": False}
                    },
                    {
                        'type': "bubble_text",
                        "text": "txt",
                        "markdown": False
                    },
                    {
                        'type': "item_card",
                        "text": "card1",
                        "requirement": {"type": "test", "result": True}
                    }
                ],
            "root":
                [
                    {
                        'type': "pronounce_text",
                        "text": "pron",
                    }
                ],
            "suggestions":
                [
                    {
                        "type": "suggest_text",
                        "title": "pron",
                        "text": "txt",
                    },
                    {
                        "type": "suggest_text",
                        "title": "pron",
                        "text": "txt",
                        "requirement": {"type": "test", "result": True}
                    },
                    {
                        "type": "suggest_deeplink",
                        "title": "pron",
                        "deep_link": "dl"
                    }
                ]
        }
        exp1 = json.dumps({'messageName': 'ANSWER_TO_USER', 'payload': {'items': [{'bubble': {'text': 't2', 'markdown': False}}, {'card': {'cards_params': 'a lot of params'}}], 'suggestions': {'buttons': [{'title': 'p2', 'action': {'text': 't2', 'type': 'text'}}, {'title': 'p2', 'action': {'text': 't2', 'type': 'text'}}, {'title': 'p2', 'action': {'deep_link': 'www.ww.w', 'type': 'deep_link'}}]}, 'pronounceText': 'p2'}}, sort_keys=True)
        exp2 = json.dumps({'messageName': 'ANSWER_TO_USER', 'payload': {'items': [{'bubble': {'text': 'Ivan Ivanov', 'markdown': False}}, {'card': {'cards_params': 'a lot of params'}}], 'suggestions': {'buttons': [{'title': 'p1', 'action': {'text': 'Ivan Ivanov', 'type': 'text'}}, {'title': 'p1', 'action': {'text': 'Ivan Ivanov', 'type': 'text'}}, {'title': 'p1', 'action': {'deep_link': 'www.ww.w', 'type': 'deep_link'}}]}, 'pronounceText': 'p1'}}, sort_keys=True)

        action = SDKAnswerToUser(items)
        for i in range(3):
            result = action.run(user, None)
            self.assertTrue(json.dumps(result[0].raw, sort_keys=True) in [exp1, exp2])

    def test_SDKItemAnswer_root(self):

        registered_factories[SdkAnswerItem] = items_factory
        answer_items["bubble_text"] = BubbleText
        answer_items["item_card"] = ItemCard
        answer_items["pronounce_text"] = PronounceText
        answer_items["suggest_text"] = SuggestText
        answer_items["suggest_deeplink"] = SuggestDeepLink


        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        user.message = PicklableMock()
        user.message.payload = {"personInfo": {"name": "Ivan Ivanov"}}
        items = {
            "type": "sdk_answer_to_user",
            "static":
                {
                    "static_text": "st1",
                    "card1": {"cards_params": "a lot of params"},
                    "dl": "www.ww.w"
                },
            "random_choice": [
                {
                    "pron": "p1",
                    "txt": "{{payload.personInfo.name}}",
                    "title": "title1"
                },
                {
                    "pron": "p2",
                    "txt": "t2",
                    "title": "title2"
                }],
            "root":
                [
                    {
                        'type': "pronounce_text",
                        "text": "pron",
                    },
                ]
        }
        exp1 = json.dumps({'messageName': 'ANSWER_TO_USER', 'payload': {'pronounceText': 'p1'}}, sort_keys=True)
        exp2 = json.dumps({'messageName': 'ANSWER_TO_USER', 'payload': {'pronounceText': 'p2'}}, sort_keys=True)

        action = SDKAnswerToUser(items)
        for i in range(3):
            result = action.run(user, None)
            self.assertTrue(json.dumps(result[0].raw, sort_keys=True) in [exp1, exp2])

    def test_SDKItemAnswer_simple(self):

        registered_factories[SdkAnswerItem] = items_factory
        answer_items["bubble_text"] = BubbleText

        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        items = {
            "type": "sdk_answer_to_user",
            "items":
                [
                    {
                        'type': "bubble_text",
                        "text": "42"
                    }
                ]
        }
        action = SDKAnswerToUser(items)
        result = action.run(user, None)
        self.assertDictEqual(result[0].raw, {'messageName': 'ANSWER_TO_USER', 'payload': {'items': [{'bubble': {'text': '42', 'markdown': True}}]}})

    def test_SDKItemAnswer_suggestions_template(self):

        registered_factories[SdkAnswerItem] = items_factory
        answer_items["bubble_text"] = BubbleText

        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        items = {
            "type": "sdk_answer_to_user",
            "support_templates": {
                "suggestions_from_template": '{ "buttons": [ { "title": "some title", "action": { "type": "text", "text": "some text" } } ]}'
            },
            "suggestions_template": {
                "type": "unified_template",
                "template": "{{ suggestions_from_template }}",
                "loader": "json"
            }
        }
        action = SDKAnswerToUser(items)
        result = action.run(user, None)
        self.assertDictEqual(
            result[0].raw,
            {
                'messageName': 'ANSWER_TO_USER',
                'payload': {
                    'suggestions': {
                        'buttons': [
                            {'title': 'some title', 'action': {'type': 'text', 'text': 'some text'}}
                        ]
                    }
                }
            })


class GiveMeMemoryActionTest(unittest.TestCase):
    @patch("smart_kit.configs.settings.Settings")
    def test_run(self, settings_mock: MagicMock):
        expected = [
            Command("GIVE_ME_MEMORY",
                    {
                        'root_nodes': {
                            'protocolVersion': 1
                        },
                        'consumer': {
                            'projectId': '0'
                        },
                        "tokenType": 0,
                        'profileEmployee': 0,
                        'memory': [
                            {
                                'memoryPartition': 'confidentialMemo',
                                'tags': [
                                    'userAgreement',
                                    'userAgreementProject'
                                ]
                            },
                            {
                                'memoryPartition': 'projectPrivateMemo',
                                'tags': [
                                    'test'
                                ]
                            }
                        ]
                    },
                    None, "kafka", {"topic_key": "client_info", "kafka_key": "main", "kafka_replyTopic": "app"})
        ]
        user = PicklableMagicMock()
        params = {"params": "params"}
        user.parametrizer = MockSimpleParametrizer(user, {"data": params})
        settings = {
            "template_settings": {
                "project_id": "0",
                "consumer_topic": "app"
            }
        }
        settings_mock.return_value = settings
        items = {
            "behavior": "my_behavior",
            "nodes": {
                "memory": {
                    "confidentialMemo": [
                        "userAgreement",
                        "userAgreementProject"
                    ],
                    "projectPrivateMemo": [
                        "{{ 'test' }}"
                    ]
                },
                "profileEmployee": {
                    "type": "unified_template",
                    "template": "{{ 0 }}",
                    "loader": "json"
                },
                "tokenType": {
                    "type": "unified_template",
                    "template": "{{ 0 }}",
                    "loader": "json"
                }
            }
        }
        text_preprocessing_result = PicklableMock()
        action = GiveMeMemoryAction(items)
        result = action.run(user, text_preprocessing_result)
        self.assertEqual(expected[0].name, result[0].name)
        self.assertEqual(expected[0].payload, result[0].payload)


class RememberThisActionTest(unittest.TestCase):
    def test_run(self):
        expected = [
            Command("REMEMBER_THIS",
                    {
                        'root_nodes': {
                            'protocolVersion': 3
                        },
                        'consumer': {
                            'projectId': '0'
                        },
                        'clientIds': 0,
                        'memory': [
                            {
                                'memoryPartition': 'publicMemo',
                                'partitionData': [
                                    {
                                        'tag': 'historyInfo',
                                        'action': {
                                            'type': 'upsert',
                                            'params': {
                                                'operation': [
                                                    {
                                                        'selector': {
                                                            'intent': {
                                                                '$eq': 'run_app'
                                                            },
                                                            'surface': {
                                                                '$eq': '0'
                                                            },
                                                            'channel': {
                                                                '$eq': '0'
                                                            },
                                                            'projectId': {
                                                                '$eq': '0'
                                                            }
                                                        },
                                                        'updater': [
                                                            {
                                                                '$set': {
                                                                    '$.lastExecuteDateTime': '0'
                                                                }
                                                            },
                                                            {
                                                                '$inc': {
                                                                    '$.executeCounter': 1
                                                                }
                                                            }
                                                        ]
                                                    }
                                                ]
                                            }
                                        }
                                    }
                                ]
                            }
                        ]
                    },
                    None, "kafka", {
                        "topic_key": 'client_info_remember',
                        "kafka_key": "main",
                        "kafka_replyTopic": "app"
                    })
        ]
        user = PicklableMagicMock()
        user.settings = {
            "template_settings": {
                "project_id": "0",
                "consumer_topic": "app"
            }
        }
        params = {"params": "params"}
        user.parametrizer = MockSimpleParametrizer(user, {"data": params})
        items = {
            "nodes": {
              "clientIds": {
                "type": "unified_template",
                "template": "{{ 0 }}",
                "loader": "json"
              },
              "memory": [
                {
                  "memoryPartition": "publicMemo",
                  "partitionData": [
                    {
                      "tag": "historyInfo",
                      "action": {
                        "type": "upsert",
                        "params": {
                          "operation": [
                            {
                              "selector": {
                                "intent": {
                                  "$eq": "run_app"
                                },
                                "surface": {
                                  "$eq": "{{ 0 }}"
                                },
                                "channel": {
                                  "$eq": "{{ 0 }}"
                                },
                                "projectId": {
                                  "$eq": "{{ 0 }}"
                                }
                              },
                              "updater": [
                                {
                                  "$set": {
                                    "$.lastExecuteDateTime": "{{ 0 }}"
                                  }
                                },
                                {
                                  "$inc": {
                                    "$.executeCounter": 1
                                  }
                                }
                              ]
                            }
                          ]
                        }
                      }
                    }
                  ]
                }
              ]
            }
        }
        action = RememberThisAction(items)
        result = action.run(user, None)
        self.assertEqual(expected[0].name, result[0].name)
        self.assertEqual(expected[0].payload, result[0].payload)
